# Generated by Django 5.1.7 on 2025-03-21 11:35

import aleksis.core.mixins
import colorfield.fields
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('core', '0077_add_address_to_organisation'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Role',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('managed_by_app_label', models.CharField(blank=True, editable=False, max_length=255, verbose_name='App label of app responsible for managing this instance')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
                ('name', models.CharField(max_length=255, unique=True, verbose_name='Name')),
                ('short_name', models.CharField(max_length=255, unique=True, verbose_name='Short name')),
                ('reciprocal_name', models.CharField(max_length=255, unique=True, verbose_name='Reciprocal name')),
                ('fg_color', colorfield.fields.ColorField(blank=True, default='', image_field=None, max_length=25, samples=None, verbose_name='Foreground color')),
                ('bg_color', colorfield.fields.ColorField(blank=True, default='', image_field=None, max_length=25, samples=None, verbose_name='Background color')),
            ],
            options={
                'verbose_name': 'Role',
                'verbose_name_plural': 'Roles',
            },
        ),
        migrations.AddField(
            model_name='grouptype',
            name='available_roles',
            field=models.ManyToManyField(blank=True, related_name='group_types', to='core.role', verbose_name='Available roles'),
        ),
        migrations.AddField(
            model_name='persongroupthrough',
            name='roles',
            field=models.ManyToManyField(blank=True, related_name='+', to='core.role', verbose_name='Roles'),
        ),
    ]



# Generated by Django 5.2.1 on 2025-05-18 10:30

import aleksis.core.mixins
import ckeditor.fields
import colorfield.fields
import django.contrib.postgres.fields
import django.core.validators
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0091_add_ical_vcard_role'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Todo',
            fields=[
                ('calendarevent_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.calendarevent')),
                ('percent_complete', models.PositiveSmallIntegerField(null=True, validators=[django.core.validators.MaxValueValidator(100)], verbose_name='Percent completed')),
                ('completed', models.DateTimeField(blank=True, null=True, verbose_name='Completed on')),
                ('title', models.CharField(max_length=255, verbose_name='Title')),
                ('description', models.TextField(blank=True, verbose_name='Description')),
                ('location', models.CharField(blank=True, max_length=255, verbose_name='Location')),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='sub_todos', to='core.todo', verbose_name='Parent todo')),
            ],
            options={
                'abstract': False,
                'base_manager_name': 'objects',
            },
            bases=('core.calendarevent',),
        ),
        migrations.CreateModel(
            name='PersonalTodo',
            fields=[
                ('todo_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.todo')),
                ('groups', models.ManyToManyField(blank=True, related_name='+', to='core.group')),
                ('owner', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='owned_personal_todos', to='core.person', verbose_name='Owner')),
                ('persons', models.ManyToManyField(blank=True, related_name='+', to='core.person')),
            ],
            options={
                'abstract': False,
                'base_manager_name': 'objects',
            },
            bases=('core.todo',),
        ),
    ]



# Generated by Django 5.0.7 on 2024-07-16 10:04

import django.db.models.deletion
import django_countries.fields
from django.conf import settings
from django.db import migrations, models

def migrate_addresses(apps, schema_editor):
    Person = apps.get_model("core", "Person")
    Address = apps.get_model("core", "Address")
    AddressType = apps.get_model("core", "AddressType")

    address_type = AddressType.objects.create(name="default", managed_by_app_label="core")

    for p in Person.objects.all():
        if p.street or p.housenumber or p.postal_code or p.place:
            address = Address.objects.create(
                street=p.street,
                housenumber=p.housenumber,
                postal_code=p.postal_code,
                place=p.place,
            )
            address.address_types.add(address_type)

            p.addresses.add(address)

class Migration(migrations.Migration):

    dependencies = [
        ("core", "0074_rename_sex_choices"),
    ]

    operations = [
        migrations.CreateModel(
            name="AddressType",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "managed_by_app_label",
                    models.CharField(
                        blank=True,
                        editable=False,
                        max_length=255,
                        verbose_name="App label of app responsible for managing this instance",
                    ),
                ),
                ("extended_data", models.JSONField(default=dict, editable=False)),
                ("name", models.CharField(max_length=255, verbose_name="Name")),
            ],
            options={
                "verbose_name": "Address Type",
                "verbose_name_plural": "Address Types",
            },
        ),
        migrations.CreateModel(
            name="Address",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "managed_by_app_label",
                    models.CharField(
                        blank=True,
                        editable=False,
                        max_length=255,
                        verbose_name="App label of app responsible for managing this instance",
                    ),
                ),
                ("extended_data", models.JSONField(default=dict, editable=False)),
                (
                    "street",
                    models.CharField(blank=True, max_length=255, verbose_name="Street"),
                ),
                (
                    "housenumber",
                    models.CharField(
                        blank=True, max_length=255, verbose_name="Street number"
                    ),
                ),
                (
                    "postal_code",
                    models.CharField(
                        blank=True, max_length=255, verbose_name="Postal code"
                    ),
                ),
                (
                    "place",
                    models.CharField(blank=True, max_length=255, verbose_name="Place"),
                ),
                ("country", django_countries.fields.CountryField(max_length=2, verbose_name="Country", blank=True)),
                ("address_types", models.ManyToManyField(to="core.addresstype")),
            ],
            options={
                "verbose_name": "Address",
                "verbose_name_plural": "Addresses",
            },
        ),
        migrations.CreateModel(
            name="PersonAddressThrough",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "managed_by_app_label",
                    models.CharField(
                        blank=True,
                        editable=False,
                        max_length=255,
                        verbose_name="App label of app responsible for managing this instance",
                    ),
                ),
                ("extended_data", models.JSONField(default=dict, editable=False)),
                (
                    "address",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="core.address"
                    ),
                ),
                (
                    "person",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="core.person"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="person",
            name="addresses",
            field=models.ManyToManyField(
                blank=True,
                related_name="persons",
                through="core.PersonAddressThrough",
                to="core.address",
                verbose_name="Addresses",
            ),
        ),
        migrations.RunPython(migrate_addresses),
    ]



# Generated by Django 5.2.1 on 2025-05-23 09:15

import aleksis.core.mixins
import aleksis.core.util.model_helpers
import ckeditor.fields
import colorfield.fields
import django.contrib.postgres.fields
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0094_add_dates_to_relationships'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        # this is OK as Django won't execute this step due to managed=False
        migrations.CreateModel(
            name='BirthdayEvent',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'db_table': 'core_birthdayevent',
                'managed': False,
            },
            bases=(aleksis.core.mixins.CalendarEventMixin, models.Model),
        ),
        migrations.AlterModelOptions(
            name='freebusy',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='group',
            options={'ordering': ['short_name', 'name'], 'permissions': (('assign_child_groups_to_groups', 'Can assign child groups to groups'), ('view_group_stats', 'Can view statistics about group.'), ('view_group_availability_feed', 'Can view groups aggregated availability feed')), 'verbose_name': 'Group', 'verbose_name_plural': 'Groups'},
        ),
        migrations.AlterModelOptions(
            name='holiday',
            options={'permissions': [('view_holiday_calendar', 'Can view holiday calendar')], 'verbose_name': 'Holiday', 'verbose_name_plural': 'Holidays'},
        ),
        migrations.AlterModelOptions(
            name='person',
            options={'ordering': ['last_name', 'first_name'], 'permissions': (('view_addresses', 'Can view addresses'), ('view_contact_details', 'Can view contact details'), ('view_photo', 'Can view photo'), ('view_avatar', 'Can view avatar image'), ('view_person_groups', 'Can view persons groups'), ('view_personal_details', 'Can view personal details'), ('view_person_availability_feed', 'Can view persons aggregated availability feed'), ('change_user_password', 'Can change the password of the linked user'), ('reset_user_password', 'Can send a password reset request for the linked user')), 'verbose_name': 'Person', 'verbose_name_plural': 'Persons'},
        ),
        migrations.AlterModelOptions(
            name='personalevent',
            options={'base_manager_name': 'objects'},
        ),
        migrations.AlterModelOptions(
            name='schoolterm',
            options={'ordering': ['-date_start'], 'verbose_name': 'School term', 'verbose_name_plural': 'School terms'},
        ),
        migrations.AlterField(
            model_name='availabilitytype',
            name='color',
            field=colorfield.fields.ColorField(blank=True, default=None, image_field=None, max_length=25, null=True, samples=None, verbose_name='Color'),
        ),
        migrations.AlterField(
            model_name='availabilitytype',
            name='uuid',
            field=models.UUIDField(db_default=models.Func(function='gen_random_uuid'), editable=False, unique=True),
        ),
        migrations.AlterField(
            model_name='calendarwidget',
            name='selected_calendars',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), blank=True, default=list, help_text='If no specific calendars are selected, all calendars are allowed.', size=None, verbose_name='Allowed calendars'),
        ),
        migrations.AlterField(
            model_name='calendarwidgetconfiguration',
            name='selected_calendars',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), blank=True, default=list, size=None, verbose_name='Selected calendars'),
        ),
        migrations.AlterField(
            model_name='custommenuitem',
            name='icon',
            field=models.CharField(blank=True, choices=aleksis.core.util.model_helpers.get_icon_choices, max_length=50, verbose_name='Icon'),
        ),
        migrations.AlterField(
            model_name='datacheckresult',
            name='data_check',
            field=models.CharField(choices=[('broken_dashboard_widgets', 'Ensure that there are no broken DashboardWidgets.'), ('disallowed_uid', 'Ensure that there are no disallowed usernames.'), ('field_validation_custommenuitem_icon', 'Validate field icon of model core.CustomMenuItem.'), ('email_unique', 'Ensure that email addresses are unique among all persons'), ('colors_accessibility_datacheck', 'Validate contrast accessibility of colors of customizable objects.'), ('accessible_themes_colors_datacheck', 'Validate that theme colors are accessible.')], max_length=255, verbose_name='Related data check task'),
        ),
        migrations.AlterField(
            model_name='externallinkwidget',
            name='icon_url',
            field=models.URLField(blank=True, verbose_name='Icon URL'),
        ),
        migrations.AlterField(
            model_name='externallinkwidget',
            name='url',
            field=models.URLField(blank=True, verbose_name='URL'),
        ),
        migrations.AlterField(
            model_name='notification',
            name='icon',
            field=models.CharField(choices=aleksis.core.util.model_helpers.get_icon_choices, default='information-outline', max_length=50, verbose_name='Icon'),
        ),
        migrations.AlterField(
            model_name='personinvitation',
            name='inviter',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='inviter'),
        ),
        migrations.AlterField(
            model_name='staticcontentwidget',
            name='content',
            field=ckeditor.fields.RichTextField(blank=True, verbose_name='Content'),
        ),
    ]



import base64
import json
from datetime import date

from django.contrib.auth import get_user_model
from django.test import override_settings
from django.urls import reverse

import pytest
from graphene_django.utils.testing import graphql_query
from graphql import GraphQLError

from aleksis.core.data_checks import ColorAccessibilityDataCheck
from aleksis.core.models import DataCheckResult, Group, Holiday, OAuthApplication, Person

pytestmark = pytest.mark.django_db


def test_all_settigns_registered():
    """Tests for regressions of preferences not being registered.

    https://edugit.org/AlekSIS/official/AlekSIS-Core/-/issues/592
    """

    from dynamic_preferences.types import BasePreferenceType

    from aleksis.core import preferences
    from aleksis.core.preferences import person_preferences_registry, site_preferences_registry

    for obj in preferences.__dict__.values():
        if not isinstance(obj, BasePreferenceType):
            continue

        in_site_reg = site_preferences_registry.get(obj.section.name, {}).get(obj.name, None) is obj
        in_person_reg = (
            person_preferences_registry.get(obj.section.name, {}).get(obj.name, None) is obj
        )

        assert in_site_reg != in_person_reg


def test_reassign_user_to_person():
    """Tests that on re-assigning a user, groups are correctly synced.

    https://edugit.org/AlekSIS/official/AlekSIS-Core/-/issues/628
    """

    User = get_user_model()

    group1 = Group.objects.create(name="Group 1")
    group2 = Group.objects.create(name="Group 2")

    user1 = User.objects.create(username="user1")
    user2 = User.objects.create(username="user2")

    person1 = Person.objects.create(first_name="Person", last_name="1", user=user1)
    person2 = Person.objects.create(first_name="Person", last_name="2", user=user2)

    person1.member_of.set([group1])
    person2.member_of.set([group2])

    assert user1.groups.count() == 1
    assert user2.groups.count() == 1
    assert user1.groups.first().name == "Group 1"
    assert user2.groups.first().name == "Group 2"

    person1.user = None
    person1.save()
    assert user1.groups.count() == 0

    person2.user = user1
    person2.save()
    person1.user = user2
    person1.save()

    assert user1.groups.count() == 1
    assert user2.groups.count() == 1
    assert user1.groups.first().name == "Group 2"
    assert user2.groups.first().name == "Group 1"


@override_settings(ROOT_URLCONF="aleksis.core.tests.regression.view_oauth")
def test_no_access_oauth2_client_credentials_without_allowed_scopes(client):
    """Tests that ClientProtectedResourceMixin doesn't allow access if no allowed scopes are set.

    https://edugit.org/AlekSIS/official/AlekSIS-Core/-/issues/688
    """

    wrong_application = OAuthApplication(
        name="Test Application",
        allowed_scopes=[],
        authorization_grant_type=OAuthApplication.GRANT_CLIENT_CREDENTIALS,
        client_type=OAuthApplication.CLIENT_CONFIDENTIAL,
        redirect_uris=["http://localhost:8000/"],
    )
    wrong_application_secret = wrong_application.client_secret
    wrong_application.save()
    wrong_application_2 = OAuthApplication(
        name="Test Application",
        allowed_scopes=["read"],
        authorization_grant_type=OAuthApplication.GRANT_CLIENT_CREDENTIALS,
        client_type=OAuthApplication.CLIENT_CONFIDENTIAL,
        redirect_uris=["http://localhost:8000/"],
    )
    wrong_application_2_secret = wrong_application_2.client_secret
    wrong_application_2.save()
    correct_application = OAuthApplication(
        name="Test Application",
        allowed_scopes=["write"],
        authorization_grant_type=OAuthApplication.GRANT_CLIENT_CREDENTIALS,
        client_type=OAuthApplication.CLIENT_CONFIDENTIAL,
        redirect_uris=["http://localhost:8000/"],
    )
    correct_application_secret = correct_application.client_secret
    correct_application.save()

    url = reverse("client_protected_resource_mixin_test")
    auth_header = (
        "Basic "
        + base64.b64encode(
            f"{wrong_application.client_id}:{wrong_application_secret}".encode()
        ).decode()
    )
    r = client.get(url, HTTP_AUTHORIZATION=auth_header)
    assert r.status_code == 403

    auth_header = (
        "Basic "
        + base64.b64encode(
            f"{wrong_application_2.client_id}:{wrong_application_2_secret}".encode()
        ).decode()
    )
    r = client.get(url, HTTP_AUTHORIZATION=auth_header)
    assert r.status_code == 403

    auth_header = (
        "Basic "
        + base64.b64encode(
            f"{correct_application.client_id}:{correct_application_secret}".encode()
        ).decode()
    )
    r = client.get(url, HTTP_AUTHORIZATION=auth_header)
    assert r.status_code == 200


def test_nested_queries_max_depth(logged_in_client):
    """Tests that queries enforces a maximum depth limit.

    https://edugit.org/AlekSIS/official/AlekSIS-Core/-/issues/1195
    """

    with pytest.raises(GraphQLError) as excinfo:
        graphql_query(
            """
            query person{
                personByIdOrMe{
                    id
                    memberOf{
                        id
                        members{
                            id
                            memberOf{
                                id
                                members{
                                    id
                                    memberOf{
                                        id
                                        members{
                                            id
                                            memberOf{
                                                id
                                                members{
                                                    id
                                                    memberOf{
                                                        id
                                                        members{
                                                            id
                                                            memberOf{
                                                                id
                                                                members{
                                                                    id
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            """,
            client=logged_in_client,
        )

    assert "exceeds maximum operation depth" in str(excinfo.value)

    response = graphql_query(
        """
        query person{
            personByIdOrMe{
                id
                memberOf{
                    id
                    members{
                        id
                        memberOf{
                            id
                            members{
                                id
                                memberOf{
                                    id
                                    members{
                                        id
                                        memberOf{
                                            id
                                            members{
                                                id
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        """,
        client=logged_in_client,
    )

    content = json.loads(response.content)
    assert "errors" not in content


def test_all_day_events_end_date():
    """Tests that CalendarEvent's value method returns a non-inclusive end date for all day events.

    https://edugit.org/AlekSIS/official/AlekSIS-Core/-/issues/1199
    """
    holiday = Holiday.objects.create(
        date_start=date(2024, 2, 1), date_end=date(2024, 2, 1), holiday_name="Test Holiday"
    )

    assert holiday.value_start_datetime(holiday) == date(2024, 2, 1)
    assert holiday.value_end_datetime(holiday) == date(2024, 2, 2)


def test_data_checks_delete_old_results():
    """Test that data checks run correctly with deleting old results.

    https://edugit.org/AlekSIS/official/AlekSIS-Core/-/issues/1293
    """

    g = Group.objects.create(name="Foo", short_name="foo")
    ColorAccessibilityDataCheck.register_result(g)
    ColorAccessibilityDataCheck._current_results = []

    assert DataCheckResult.objects.count() == 1

    ColorAccessibilityDataCheck.run_check_data()

    assert DataCheckResult.objects.count() == 0



# Generated by Django 5.2.1 on 2025-05-18 18:19

import aleksis.core.mixins
import ckeditor.fields
import colorfield.fields
import django.contrib.postgres.fields
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


def adapt_guardian_role(apps, schema_editor):
    RoleModel = apps.get_model('core', 'Role')

    db_alias = schema_editor.connection.alias

    guardian_role, _ = RoleModel.objects.get_or_create(
        managed_by_app_label="core",
        short_name="guardian",
        defaults={"name": "Guardian", "reciprocal_name": "Child"},
    )

    guardian_role.vcard_related_type = "parent"
    guardian_role.save()


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0090_password_perms'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='role',
            name='ical_participation_role',
            field=models.CharField(blank=True, max_length=255, verbose_name='iCal name'),
        ),
        migrations.AddField(
            model_name='role',
            name='vcard_related_type',
            field=models.CharField(blank=True, max_length=255, verbose_name='vCard name'),
        ),
        migrations.RunPython(adapt_guardian_role)
    ]



from datetime import datetime, timezone
from zoneinfo import ZoneInfo

import pytest
from recurrence import WEEKLY, Recurrence, Rule

from aleksis.core.models import CalendarEvent, Person, PersonalEvent

pytestmark = pytest.mark.django_db


def test_calendar_event_timezone():
    datetime_start = datetime(2024, 1, 1, 0, 0, tzinfo=timezone.utc).astimezone(
        ZoneInfo("Europe/Berlin")
    )
    datetime_end = datetime(2024, 12, 31, 0, 0, tzinfo=timezone.utc).astimezone(
        ZoneInfo("Europe/Berlin")
    )

    # No timezone set
    calendar_event = CalendarEvent.objects.create(
        datetime_start=datetime_start, datetime_end=datetime_end
    )
    calendar_event.refresh_from_db()

    assert calendar_event.datetime_start == datetime(2024, 1, 1, 0, 0, tzinfo=timezone.utc)
    assert calendar_event.datetime_end == datetime(2024, 12, 31, 0, 0, tzinfo=timezone.utc)
    assert CalendarEvent.value_start_datetime(calendar_event) == datetime(
        2024, 1, 1, 0, 0, tzinfo=timezone.utc
    )
    assert CalendarEvent.value_end_datetime(calendar_event) == datetime(
        2024, 12, 31, 0, 0, tzinfo=timezone.utc
    )
    assert calendar_event.timezone is None

    # Set timezone if not allowed
    calendar_event.timezone = ZoneInfo("Europe/Berlin")
    calendar_event.save()
    calendar_event.refresh_from_db()
    assert calendar_event.timezone is None

    # Automatically set timezone
    calendar_event.datetime_start = datetime_start
    calendar_event.datetime_end = datetime_end
    calendar_event.recurrences = Recurrence(dtstart=datetime_start, rrules=[Rule(WEEKLY)])
    calendar_event.save()
    calendar_event.refresh_from_db()

    assert calendar_event.datetime_start == datetime(2024, 1, 1, 0, 0, tzinfo=timezone.utc)
    assert calendar_event.datetime_end == datetime(2024, 12, 31, 0, 0, tzinfo=timezone.utc)
    assert CalendarEvent.value_start_datetime(calendar_event) == datetime_start
    assert CalendarEvent.value_end_datetime(calendar_event) == datetime_end
    assert calendar_event.timezone == ZoneInfo("Europe/Berlin")

    # Manually set timezone (e. g. from frontend)
    calendar_event.timezone = ZoneInfo("Europe/Berlin")
    calendar_event.save()
    calendar_event.refresh_from_db()
    assert calendar_event.timezone == ZoneInfo("Europe/Berlin")


def test_calendar_event_recurring_timezones_all_objects():
    tz = ZoneInfo("Europe/Berlin")
    owner = Person.objects.create(first_name="Foo", last_name="Baz")
    start_dt = datetime(2025, 3, 26, 18, 0, tzinfo=tz)
    end_dt = datetime(2025, 3, 26, 19, 0, tzinfo=tz)
    end_rrule = datetime(2025, 4, 2, 19, 0, tzinfo=tz)

    event = PersonalEvent()
    event.datetime_start = start_dt
    event.datetime_end = end_dt
    event.recurrences = Recurrence(dtstart=start_dt, rrules=[Rule(WEEKLY, until=end_rrule)])
    event.owner = owner
    event.title = "Test Event"
    event.save()

    events = sorted(
        event.get_objects(expand=True, start=datetime(2025, 1, 1), end=datetime(2025, 12, 31)),
        key=lambda event: event.odatetime,
    )
    assert len(events) == 2
    obj1 = events[0]
    obj2 = events[1]

    assert obj1.value_start_datetime(obj1).tzinfo == tz
    assert obj1.value_start_datetime(obj1) == start_dt

    assert obj2.value_start_datetime(obj2).tzinfo == tz
    assert obj2.value_start_datetime(obj2) == datetime(2025, 4, 2, 18, 0, tzinfo=tz)


def test_calendar_event_recurring_timezones_in_range():
    tz = ZoneInfo("Europe/Berlin")
    owner = Person.objects.create(first_name="Foo", last_name="Baz")
    start_dt = datetime(2025, 3, 26, 18, 0, tzinfo=tz)
    end_dt = datetime(2025, 3, 26, 19, 0, tzinfo=tz)
    end_rrule = datetime(2025, 4, 2, 19, 0, tzinfo=tz)

    event = PersonalEvent()
    event.datetime_start = start_dt
    event.datetime_end = end_dt
    event.recurrences = Recurrence(dtstart=start_dt, rrules=[Rule(WEEKLY, until=end_rrule)])
    event.owner = owner
    event.title = "Test Event"
    event.save()

    # Exact range
    events = event.get_objects(start=start_dt, end=end_dt, expand=True)
    assert events.count() == 1
    obj1 = events[0]
    assert obj1.value_start_datetime(obj1).tzinfo == tz
    assert obj1.value_start_datetime(obj1) == start_dt

    # Exact range with second occurrence (shifted because of DST)
    events = event.get_objects(
        start=datetime(2025, 4, 2, 18, 0, tzinfo=tz),
        end=datetime(2025, 4, 2, 19, 0, tzinfo=tz),
        expand=True,
    )
    assert events.count() == 1
    obj1 = events[0]
    assert obj1.value_start_datetime(obj1).tzinfo == tz
    assert obj1.value_start_datetime(obj1) == datetime(2025, 4, 2, 18, 0, tzinfo=tz)

    # Extended range with second occurrence
    events = event.get_objects(
        start=datetime(2025, 4, 2, 17, 0, tzinfo=tz),
        end=datetime(2025, 4, 2, 20, 0, tzinfo=tz),
        expand=True,
    )
    assert events.count() == 1
    obj1 = events[0]
    assert obj1.value_start_datetime(obj1).tzinfo == tz
    assert obj1.value_start_datetime(obj1) == datetime(2025, 4, 2, 18, 0, tzinfo=tz)

    # Partial range with first occurrence
    events = event.get_objects(
        start=datetime(2025, 3, 26, 18, 0, tzinfo=tz),
        end=datetime(2025, 3, 26, 18, 30, tzinfo=tz),
        expand=True,
    )
    assert events.count() == 1
    obj1 = events[0]
    assert obj1.value_start_datetime(obj1).tzinfo == tz
    assert obj1.value_start_datetime(obj1) == start_dt

    # Partial range with second occurrence
    events = event.get_objects(
        start=datetime(2025, 4, 2, 18, 0, tzinfo=tz),
        end=datetime(2025, 4, 2, 18, 30, tzinfo=tz),
        expand=True,
    )
    assert events.count() == 1
    obj1 = events[0]
    assert obj1.value_start_datetime(obj1).tzinfo == tz
    assert obj1.value_start_datetime(obj1) == datetime(2025, 4, 2, 18, 0, tzinfo=tz)


def test_calendar_event_recurring_timezones_not_in_range():
    tz = ZoneInfo("Europe/Berlin")
    owner = Person.objects.create(first_name="Foo", last_name="Baz")
    start_dt = datetime(2025, 3, 26, 18, 0, tzinfo=tz)
    end_dt = datetime(2025, 3, 26, 19, 0, tzinfo=tz)
    end_rrule = datetime(2025, 4, 2, 19, 0, tzinfo=tz)

    event = PersonalEvent()
    event.datetime_start = start_dt
    event.datetime_end = end_dt
    event.recurrences = Recurrence(dtstart=start_dt, rrules=[Rule(WEEKLY, until=end_rrule)])
    event.owner = owner
    event.title = "Test Event"
    event.save()

    events = event.get_objects(
        start=datetime(2025, 4, 2, 19, 0, tzinfo=tz),
        end=datetime(2025, 4, 2, 20, 0, tzinfo=tz),
        expand=True,
    )
    assert events.count() == 0



# Generated by Django 4.2.11 on 2024-05-08 08:55

from django.db import migrations, models



def _migrate_recurrences(apps, schema_editor):
    CalendarEvent = apps.get_model('core', 'CalendarEvent')
    for event in CalendarEvent.objects.using(schema_editor.connection.alias).all():
        if not event.recurrences:
            continue
        rule = event.recurrences.rrules[0]
        event.rrule_until = rule.until
        event.save()

class Migration(migrations.Migration):

    dependencies = [
        ('core', '0066_group_type_owner_permissions'),
    ]

    operations = [
        migrations.AddField(
            model_name='calendarevent',
            name='rrule_until',
            field=models.DateTimeField(blank=True, editable=False, null=True),
        ),
        migrations.RunPython(_migrate_recurrences)
    ]



# Generated by Django 5.2.1 on 2025-05-18 15:55

import aleksis.core.mixins
import ckeditor.fields
import colorfield.fields
import django.contrib.postgres.fields
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0089_taskuserassignment_task_id'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='globalpermissions',
            options={'default_permissions': (), 'managed': False, 'permissions': (('view_system_status', 'Can view system status'), ('manage_data', 'Can manage data'), ('impersonate', 'Can impersonate'), ('search', 'Can use search'), ('change_site_preferences', 'Can change site preferences'), ('change_person_preferences', 'Can change person preferences'), ('change_group_preferences', 'Can change group preferences'), ('test_pdf', 'Can test PDF generation'), ('invite', 'Can invite persons'), ('view_birthday_calendar', 'Can view birthday calendar'), ('reset_user_password', 'Can reset passwords for users'), ('change_user_password', 'Can change passwords for users'))},
        ),
    ]



from django.contrib.contenttypes.models import ContentType

import graphene
from graphene_django import DjangoObjectType

from ..models import (
    Announcement,
    Group,
    Person,
)
from .base import (
    BaseBatchCreateMutation,
    BaseBatchDeleteMutation,
    BaseBatchPatchMutation,
    PermissionsTypeMixin,
)
from .group import GroupType
from .person import PersonType


class AnnouncementType(PermissionsTypeMixin, DjangoObjectType):
    class Meta:
        model = Announcement
        fields = (
            "id",
            "datetime_start",
            "datetime_end",
            "title",
            "description",
            "priority",
            "is_global",
        )

    recipient_groups = graphene.List(GroupType)
    recipient_persons = graphene.List(PersonType)

    def resolve_recipient_groups(root, info, **kwargs):
        return root.get_recipients_for_model(Group)

    def resolve_recipient_persons(root, info, **kwargs):
        return root.get_recipients_for_model(Person)


class DisplayAnnouncementType(DjangoObjectType):
    class Meta:
        model = Announcement
        skip_registry = True
        fields = (
            "id",
            "datetime_start",
            "datetime_end",
            "title",
            "description",
            "priority",
        )


class AnnouncementBatchCreateMutation(BaseBatchCreateMutation):
    class Meta:
        model = Announcement
        permissions = ("core.create_announcement_rule",)
        only_fields = (
            "datetime_start",
            "datetime_end",
            "title",
            "description",
            "priority",
            "is_global",
        )
        custom_fields = {
            "recipient_groups": graphene.List(graphene.ID),
            "recipient_persons": graphene.List(graphene.ID),
        }

    @classmethod
    def after_mutate(cls, root, info, input, created_objs, return_data):  # noqa
        """Create AnnouncementRecipients"""
        for spec, obj in zip(input, created_objs):
            for group in spec.recipient_groups:
                obj.recipients.create(recipient=Group.objects.get(id=group))
            for person in spec.recipient_persons:
                obj.recipients.create(recipient=Person.objects.get(id=person))


class AnnouncementBatchDeleteMutation(BaseBatchDeleteMutation):
    class Meta:
        model = Announcement
        permissions = ("core.delete_announcement_rule",)

    @classmethod
    def before_save(cls, root, info, ids, qs_to_delete):
        """Delete AnnouncementRecipients"""
        for announcement in qs_to_delete:
            # Copied from forms.py
            announcement.recipients.all().delete()


class AnnouncementBatchPatchMutation(BaseBatchPatchMutation):
    class Meta:
        model = Announcement
        permissions = ("core.edit_announcement_rule",)
        only_fields = (
            "id",
            "datetime_start",
            "datetime_end",
            "title",
            "description",
            "priority",
            "is_global",
        )
        custom_fields = {
            "recipient_groups": graphene.List(graphene.ID),
            "recipient_persons": graphene.List(graphene.ID),
        }

    @classmethod
    def after_mutate(cls, root, info, input, updated_objs, return_data):  # noqa
        """Update AnnouncementRecipients"""

        def delete_announcement_recipients_of_model(announcement, model):
            # Del & then recreate
            ct = ContentType.objects.get_for_model(model)
            announcement.recipients.filter(content_type=ct).delete()

        for spec, obj in zip(input, updated_objs):
            if spec.recipient_groups:
                delete_announcement_recipients_of_model(obj, Group)
                for group in spec.recipient_groups:
                    obj.recipients.create(recipient=Group.objects.get(id=group))

            if spec.recipient_persons:
                delete_announcement_recipients_of_model(obj, Person)
                for person in spec.recipient_persons:
                    obj.recipients.create(recipient=Person.objects.get(id=person))



import graphene
from graphene_django import DjangoObjectType

from ..models import PersonInvitation


class PersonInvitationType(DjangoObjectType):
    class Meta:
        model = PersonInvitation
        fields = [
            "id",
        ]

    valid = graphene.Boolean()
    has_email = graphene.Boolean()
    has_person = graphene.Boolean()

    @staticmethod
    def resolve_valid(root, info, **kwargs):
        return not root.accepted and not root.key_expired()

    @staticmethod
    def resolve_has_email(root, info, **kwargs):
        return bool(root.email)

    @staticmethod
    def resolve_has_person(root, info, **kwargs):
        return bool(root.person)



from datetime import date, datetime
from typing import TYPE_CHECKING, Optional, Union

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.db.models import BooleanField, Case, Model, Q, QuerySet, Value, When
from django.db.models.manager import Manager
from django.utils import timezone

from calendarweek import CalendarWeek
from django_cte.cte import CTEQuerySet
from django_pg_rrule.managers import RecurrenceManager
from guardian.shortcuts import get_objects_for_user
from polymorphic.managers import PolymorphicManager
from polymorphic.query import PolymorphicQuerySet

from .util.core_helpers import get_site_preferences, has_person

if TYPE_CHECKING:
    from .models import Group, Person, SchoolTerm


class AlekSISBaseManager(Manager):
    """Base manager for AlekSIS model customisation."""

    def unmanaged(self) -> QuerySet:
        """Get instances that are not managed by any particular app."""
        return super().get_queryset().filter(managed_by_app_label="")

    def managed_by_app(self, app_label: str) -> QuerySet:
        """Get instances managed by a particular app."""
        return super().get_queryset().filter(managed_by_app_label=app_label)


class AlekSISBaseManagerWithoutMigrations(AlekSISBaseManager):
    """AlekSISBaseManager for auto-generating managers just by query sets."""

    use_in_migrations = False


class PolymorphicBaseManager(AlekSISBaseManagerWithoutMigrations, PolymorphicManager):
    """Default manager for extensible, polymorphic models."""


class CalendarEventMixinQuerySet(CTEQuerySet):
    pass


class CalendarEventQuerySet(PolymorphicQuerySet, CalendarEventMixinQuerySet):
    pass


class CalendarEventMixinManager(RecurrenceManager):
    queryset_class = CalendarEventMixinQuerySet


class CalendarEventManager(PolymorphicBaseManager, CalendarEventMixinManager):
    queryset_class = CalendarEventQuerySet


class DateRangeQuerySetMixin:
    """QuerySet with custom query methods for models with date ranges.

    Filterable fields: date_start, date_end
    """

    def within_dates(self, start: date, end: date):
        """Filter for all objects within a date range."""
        return self.filter(date_start__lte=end, date_end__gte=start)

    def in_week(self, wanted_week: CalendarWeek):
        """Filter for all objects within a calendar week."""
        return self.within_dates(wanted_week[0], wanted_week[6])

    def on_day(self, day: date):
        """Filter for all objects on a certain day."""
        return self.within_dates(day, day)


class SchoolTermQuerySet(QuerySet, DateRangeQuerySetMixin):
    """Custom query set for school terms."""


class SchoolTermRelatedQuerySet(QuerySet):
    """Custom query set for all models related to school terms."""

    def within_dates(self, start: date, end: date) -> "SchoolTermRelatedQuerySet":
        """Filter for all objects within a date range."""
        return self.filter(school_term__date_start__lte=end, school_term__date_end__gte=start)

    def in_week(self, wanted_week: CalendarWeek) -> "SchoolTermRelatedQuerySet":
        """Filter for all objects within a calendar week."""
        return self.within_dates(wanted_week[0], wanted_week[6])

    def on_day(self, day: date) -> "SchoolTermRelatedQuerySet":
        """Filter for all objects on a certain day."""
        return self.within_dates(day, day)

    def for_school_term(self, school_term: "SchoolTerm") -> "SchoolTermRelatedQuerySet":
        return self.filter(school_term=school_term)

    def for_current_school_term_or_all(self) -> "SchoolTermRelatedQuerySet":
        """Get all objects related to current school term.

        If there is no current school term, it will return all objects.
        """
        from aleksis.core.models import SchoolTerm

        current_school_term = SchoolTerm.current
        if current_school_term:
            return self.for_school_term(current_school_term)
        else:
            return self

    def for_current_school_term_or_none(self) -> Union["SchoolTermRelatedQuerySet", None]:
        """Get all objects related to current school term.

        If there is no current school term, it will return `None`.
        """
        from aleksis.core.models import SchoolTerm

        current_school_term = SchoolTerm.current
        if current_school_term:
            return self.for_school_term(current_school_term)
        else:
            return None


class PersonManager(AlekSISBaseManagerWithoutMigrations):
    """Manager adding specific methods to persons."""


class PersonQuerySet(QuerySet):
    def annotate_permissions(self, user):
        from .models import Person

        can_edit_qs = get_objects_for_user(user, "core.change_person", self).values_list(
            "id", flat=True
        )
        if has_person(user) and get_site_preferences()["account__editable_fields_person"]:
            can_edit_qs = can_edit_qs.union(
                Person.objects.filter(id=user.person.id).values_list("id", flat=True)
            )
        can_delete_qs = get_objects_for_user(user, "core.delete_person", self).values_list(
            "id", flat=True
        )

        qs = self.annotate(
            can_edit=Case(
                When(id__in=can_edit_qs, then=Value(True)),
                default=Value(False),
                output_field=BooleanField(),
            ),
            can_delete=Case(
                When(id__in=can_delete_qs, then=Value(True)),
                default=Value(False),
                output_field=BooleanField(),
            ),
        )
        return qs


class GroupManager(AlekSISBaseManagerWithoutMigrations):
    """Manager adding specific methods to groups."""


class GroupQuerySet(SchoolTermRelatedQuerySet):
    def annotate_permissions(self, user):
        can_edit_qs = get_objects_for_user(user, "core.change_group", self).values_list(
            "id", flat=True
        )
        can_delete_qs = get_objects_for_user(user, "core.delete_group", self).values_list(
            "id", flat=True
        )

        qs = self.annotate(
            can_edit=Case(
                When(id__in=can_edit_qs, then=Value(True)),
                default=Value(False),
                output_field=BooleanField(),
            ),
            can_delete=Case(
                When(id__in=can_delete_qs, then=Value(True)),
                default=Value(False),
                output_field=BooleanField(),
            ),
        )
        return qs


class UninstallRenitentPolymorphicManager(PolymorphicManager):
    """A custom manager for django-polymorphic that filters out submodels of unavailable apps."""

    def get_queryset(self):
        DashboardWidget = apps.get_model("core", "DashboardWidget")
        if self.model is DashboardWidget:
            return super().get_queryset().instance_of(*self.model.__subclasses__())
        else:
            # Called on subclasses
            return super().get_queryset()


class DashboardWidgetInstanceQuerySet(QuerySet):
    def on(self):
        DashboardWidget = apps.get_model("core", "DashboardWidget")
        return self.filter(widget__status=DashboardWidget.Status.ON.value)

    def default_dashboard(self):
        """Get default dashboard."""

        return self.on().filter(person=None)

    def for_person(self, person):
        """Get dashboard for person."""
        return self.on().filter(person=person)


class InstalledWidgetsDashboardWidgetInstanceManager(Manager):
    """A manager that only returns DashboardWidgetInstance objects with an existing widget."""

    def get_queryset(self):
        DashboardWidget = apps.get_model("core", "DashboardWidget")

        dashboard_widget_pks = DashboardWidget.objects.all().values("id")

        return super().get_queryset().filter(widget_id__in=dashboard_widget_pks)


class HolidayQuerySet(DateRangeQuerySetMixin, CalendarEventQuerySet):
    """QuerySet with custom query methods for holidays."""

    def get_all_days(self) -> list[date]:
        """Get all days included in the selected holidays."""
        holiday_days = []
        for holiday in self:
            holiday_days += list(holiday.get_days())
        return holiday_days


class HolidayManager(CalendarEventManager):
    queryset_class = HolidayQuerySet


class AvailabilityEventQuerySet(DateRangeQuerySetMixin, CalendarEventQuerySet):
    """QuerySet with custom query methods for availability events."""

    @staticmethod
    def for_groups_q(groups: list[Union[int, "Group"]]) -> Q:
        """Get all availability events that are related to the given groups."""
        from .models import AvailabilityEvent

        pks = (
            AvailabilityEvent.objects.filter(person__member_of__in=groups)
            .values_list("pk", flat=True)
            .union(
                AvailabilityEvent.objects.filter(person__owner_of__in=groups).values_list(
                    "pk", flat=True
                )
            )
        )
        return Q(pk__in=pks)

    def for_groups(self, groups: list[Union[int, "Group"]]):
        """Get all availability events for a list of groups."""
        return self.filter(self.for_groups_q(groups)).distinct()

    @staticmethod
    def for_persons_q(persons: list[Union[int, "Person"]]) -> Q:
        """Get all availability events that are related to the given groups."""
        return Q(person__in=persons)

    def for_persons(self, persons: list["Person"]):
        """Get all availability events for a list of persons"""
        return self.filter(self.for_persons_q(persons))

    @staticmethod
    def for_groups_or_persons_q(
        persons: list[Union[int, "Person"]], groups: list[Union[int, "Group"]]
    ) -> Q:
        """Get all lesson availability events that are related to the given groups or persons."""
        from .models import AvailabilityEvent

        pks = (
            AvailabilityEvent.objects.filter(person__member_of__in=groups)
            .values_list("pk", flat=True)
            .union(
                AvailabilityEvent.objects.filter(person__owner_of__in=groups).values_list(
                    "pk", flat=True
                )
            )
            .union(
                AvailabilityEvent.objects.filter(person__in=persons).values_list("pk", flat=True)
            )
        )
        return Q(pk__in=pks)

    def for_groups_or_persons(
        self, persons: list[Union[int, "Person"]], groups: list[Union[int, "Group"]]
    ) -> "AvailabilityEventQuerySet":
        return self.filter(self.for_groups_or_persons_q(persons, groups)).distinct()


class AvailabilityEventManager(CalendarEventManager):
    queryset_class = AvailabilityEventQuerySet


class PersonalEventQuerySet(DateRangeQuerySetMixin, CalendarEventQuerySet):
    """QuerySet with custom query methods for personal events."""

    @staticmethod
    def for_persons_q(persons: list[Union[int, "Person"]]) -> Q:
        """Get all personal events that are related to the given persons."""
        from .models import PersonalEvent

        pks = (
            PersonalEvent.objects.filter(owner__in=persons)
            .values_list("pk", flat=True)
            .union(PersonalEvent.objects.filter(persons__in=persons).values_list("pk", flat=True))
            .union(
                PersonalEvent.objects.filter(groups__members__in=persons).values_list(
                    "pk", flat=True
                )
            )
        )
        return Q(pk__in=pks)

    def for_persons(self, persons: list[Union[int, "Person"]]):
        """Get all personal events for a list of persons"""
        return self.filter(self.for_persons_q(persons)).distinct()


class PersonalEventManager(CalendarEventManager):
    queryset_class = PersonalEventQuerySet


class PersonalTodoQuerySet(DateRangeQuerySetMixin, CalendarEventQuerySet):
    """QuerySet with custom query methods for personal todos."""

    @staticmethod
    def for_persons_q(persons: list[Union[int, "Person"]]) -> Q:
        """Get all personal todos that are related to the given persons."""
        from .models import PersonalTodo

        pks = (
            PersonalTodo.objects.filter(owner__in=persons)
            .values_list("pk", flat=True)
            .union(PersonalTodo.objects.filter(persons__in=persons).values_list("pk", flat=True))
            .union(
                PersonalTodo.objects.filter(groups__members__in=persons).values_list(
                    "pk", flat=True
                )
            )
        )
        return Q(pk__in=pks)

    def for_persons(self, persons: list[Union[int, "Person"]]):
        """Get all personal todos for a list of persons"""
        return self.filter(self.for_persons_q(persons)).distinct()


class PersonalTodoManager(CalendarEventManager):
    queryset_class = PersonalTodoQuerySet


class AnnouncementQuerySet(CalendarEventQuerySet):
    """Queryset for announcements providing time-based utility functions."""

    def relevant_for(self, obj: Union[Model, QuerySet]) -> QuerySet:
        """Get all relevant announcements.

        Get a QuerySet with all announcements relevant for a certain Model (e.g. a Group)
        or a set of models in a QuerySet.
        """
        if isinstance(obj, QuerySet):
            ct = ContentType.objects.get_for_model(obj.model)
            pks = list(obj.values_list("pk", flat=True))
        else:
            ct = ContentType.objects.get_for_model(obj)
            pks = [obj.pk]

        return self.filter(
            Q(recipients__content_type=ct, recipients__recipient_id__in=pks) | Q(is_global=True)
        ).distinct()

    def at_time(self, when: Optional[datetime] = None) -> QuerySet:
        """Get all announcements at a certain time."""
        when = when or timezone.now()

        # Get announcements by time
        announcements = self.filter(datetime_start__lte=when, datetime_end__gte=when)

        return announcements

    def on_date(self, when: Optional[date] = None) -> QuerySet:
        """Get all announcements at a certain date."""
        when = when or timezone.now().date()

        # Get announcements by time
        announcements = self.filter(datetime_start__date__lte=when, datetime_end__date__gte=when)

        return announcements

    def within_days(self, start: date, stop: date) -> QuerySet:
        """Get all announcements valid for a set of days."""
        # Get announcements
        announcements = self.filter(datetime_start__date__lte=stop, datetime_end__date__gte=start)

        return announcements


class AnnouncementManager(CalendarEventManager):
    queryset_class = AnnouncementQuerySet

    def relevant_for(self, obj: Union[Model, QuerySet]) -> QuerySet:
        return self.get_queryset().relevant_for(obj)

    def at_time(self, when: Optional[datetime] = None) -> QuerySet:
        return self.get_queryset().at_time(when)

    def on_date(self, when: Optional[date] = None) -> QuerySet:
        return self.get_queryset().on_date(when)

    def within_days(self, start: date, stop: date) -> QuerySet:
        return self.get_queryset().within_days(start, stop)



from uuid import uuid4

from django.conf import settings

import pytest

from aleksis.core.mixins import ExtensibleModel
from aleksis.core.models import Group, Person

pytestmark = pytest.mark.django_db


def test_get_object_uri():
    _uuid = uuid4()
    _person = Person.objects.create(first_name="Jane", last_name="Doe", uuid=_uuid)

    assert _person.get_object_uri() == f"{settings.BASE_URL}/o/core/person/{_uuid}"


def test_from_object_uri_generic():
    _uuid = uuid4()
    _person = Person.objects.create(first_name="Jane", last_name="Doe", uuid=_uuid)

    assert ExtensibleModel.from_object_uri(f"{settings.BASE_URL}/o/core/person/{_uuid}") == _person


def test_from_object_uri_concrete():
    _uuid = uuid4()
    _person = Person.objects.create(first_name="Jane", last_name="Doe", uuid=_uuid)

    assert Person.from_object_uri(f"{settings.BASE_URL}/o/core/person/{_uuid}") == _person


def test_from_object_uri_wrong_cls():
    _uuid = uuid4()
    _person = Person.objects.create(first_name="Jane", last_name="Doe", uuid=_uuid)

    assert Group.from_object_uri(f"{settings.BASE_URL}/o/core/person/{_uuid}") is None


def test_from_object_uri_no_match():
    assert Person.from_object_uri("foo") is None



from django.db.models import PositiveSmallIntegerField

from calendarweek.django import i18n_day_name_choices_lazy


class WeekdayField(PositiveSmallIntegerField):
    def __init__(self, *args, **kwargs):
        kwargs["choices"] = i18n_day_name_choices_lazy()
        super().__init__(*args, **kwargs)



from django.contrib.auth.models import Permission

import pytest

from aleksis.core.models import Group, GroupType, Person

pytestmark = pytest.mark.django_db


def test_persons_query(client_query):
    p = Person.objects.first()
    correct_group_type = GroupType.objects.create(name="correct")
    wrong_group_type = GroupType.objects.create(name="wrong")

    group_not_owner = Group.objects.create(name="not_owner")
    group_correct_group_type_owner = Group.objects.create(
        name="correct_group_type_owner", group_type=correct_group_type
    )

    group2_correct_group_type_owner = Group.objects.create(
        name="correct_group_type_owner", group_type=correct_group_type
    )
    group_wrong_group_type_owner = Group.objects.create(
        name="wrong_group_type_owner", group_type=wrong_group_type
    )
    group_no_group_type_owner = Group.objects.create(name="no_group_type_owner")

    for g in (
        group_correct_group_type_owner,
        group2_correct_group_type_owner,
        group_wrong_group_type_owner,
        group_no_group_type_owner,
    ):
        g.owners.add(p)

    correct_member = Person.objects.create(first_name="correct_member", last_name="correct_member")
    correct_member_2 = Person.objects.create(
        first_name="correct_member_2", last_name="correct_member_2"
    )
    wrong_member = Person.objects.create(first_name="wrong_member", last_name="wrong_member")

    for g in (group_correct_group_type_owner, group2_correct_group_type_owner):
        g.members.add(correct_member, correct_member_2)

    for g in (group_not_owner, group_wrong_group_type_owner, group_no_group_type_owner):
        g.members.add(wrong_member)

    response, content = client_query("{persons{id}}")
    assert len(content["data"]["persons"]) == 1
    assert content["data"]["persons"][0]["id"] == str(p.id)

    for g in Person.objects.exclude(pk=p.id):
        response, content = client_query(
            "query personById($id: ID) {object: personById(id: $id) { id } }",
            variables={"id": g.id},
        )
        assert content["data"]["object"] is None

    global_permission = Permission.objects.get(
        codename="view_person", content_type__app_label="core"
    )
    p.user.user_permissions.add(global_permission)

    response, content = client_query("{persons{id}}")
    assert set(int(g["id"]) for g in content["data"]["persons"]) == set(
        Person.objects.values_list("id", flat=True)
    )

    p.user.user_permissions.remove(global_permission)

    correct_group_type.owners_can_see_members = True
    correct_group_type.save()

    response, content = client_query("{persons{id}}")
    assert set(int(g["id"]) for g in content["data"]["persons"]) == {
        p.id,
        correct_member.id,
        correct_member_2.id,
    }

    for g in (correct_member, correct_member_2):
        response, content = client_query(
            "query personById($id: ID) {object: personById(id: $id) { id } }",
            variables={"id": g.id},
        )
        assert content["data"]["object"]["id"] == str(g.id)

    response, content = client_query(
        "query personById($id: ID) {object: personById(id: $id) { id } }",
        variables={"id": wrong_member.id},
    )
    assert content["data"]["object"] is None


def test_create_person_with_file_upload(client_query, uploaded_picture):
    p = Person.objects.first()
    global_permission = Permission.objects.get(
        codename="add_person", content_type__app_label="core"
    )
    p.user.user_permissions.add(global_permission)

    query = """
    mutation createPersons($input: [BatchCreatePersonInput]!) {
      createPersons(input: $input) {
        items: persons {
          id
          firstName
          lastName
        }
      }
    }
    """
    variables = {
        "input": [
            {
                "firstName": "Foo",
                "lastName": "Bar",
                "photo": None,
                "avatar": None,
            },
        ]
    }
    files = {
        "variables.input.0.photo": uploaded_picture,
        "variables.input.0.avatar": uploaded_picture,
    }

    response, content = client_query(query, variables=variables, files=files)
    person_id = content["data"]["createPersons"]["items"][0]["id"]

    created_person = Person.objects.get(id=person_id)
    assert created_person.first_name == "Foo"
    assert created_person.last_name == "Bar"
    assert created_person.photo.file.name.endswith(".jpg")
    assert created_person.avatar.file.name.endswith(".jpg")


def test_edit_person_with_file_upload(client_query, uploaded_picture):
    p = Person.objects.first()
    global_permission = Permission.objects.get(
        codename="change_person", content_type__app_label="core"
    )
    p.user.user_permissions.add(global_permission)

    p = Person.objects.create(
        first_name="Foo",
        last_name="Bar",
    )

    query = """
    mutation updatePersons($input: [BatchPatchPersonInput]!) {
      updatePersons(input: $input) {
        items: persons {
          id
          firstName
          lastName
        }
      }
    }
    """

    # Edit with adding files
    variables = {
        "input": [
            {
                "id": p.id,
                "firstName": "Foo",
                "lastName": "Bar",
                "photo": None,
                "avatar": None,
            },
        ]
    }
    files = {
        "variables.input.0.photo": uploaded_picture,
        "variables.input.0.avatar": uploaded_picture,
    }

    response, content = client_query(query, variables=variables, files=files)

    p.refresh_from_db()

    assert p.first_name == "Foo"
    assert p.last_name == "Bar"
    assert p.photo.file.name.endswith(".jpg")
    assert p.avatar.file.name.endswith(".jpg")

    # Edit without changing files
    variables = {
        "input": [
            {
                "id": p.id,
                "firstName": "Foo",
                "lastName": "Baz",
            },
        ]
    }
    files = {}
    response, content = client_query(query, variables=variables, files=files)
    p.refresh_from_db()

    assert p.first_name == "Foo"
    assert p.last_name == "Baz"
    assert p.photo.file.name.endswith(".jpg")
    assert p.avatar.file.name.endswith(".jpg")

    # Edit with deleting files
    variables = {
        "input": [
            {
                "id": p.id,
                "firstName": "Foo",
                "lastName": "Baz",
                "photo": None,
                "avatar": None,
            },
        ]
    }
    response, content = client_query(query, variables=variables, files=files)
    p.refresh_from_db()

    assert p.first_name == "Foo"
    assert p.last_name == "Baz"
    assert not p.photo
    assert not p.avatar


def test_create_person_with_address(client_query):
    p = Person.objects.first()
    global_permission = Permission.objects.get(
        codename="add_person", content_type__app_label="core"
    )
    p.user.user_permissions.add(global_permission)

    query = """
    mutation createPersons($input: [BatchCreatePersonInput]!) {
      createPersons(input: $input) {
        items: persons {
          id
          firstName
          lastName
        }
      }
    }
    """
    variables = {
        "input": [
            {
                "firstName": "Foo",
                "lastName": "Bar",
                "street": "Teststreet",
                "housenumber": "123",
                "postalCode": "12345",
                "place": "Testcity",
                "country": "DE",
            },
        ]
    }

    response, content = client_query(query, variables=variables)
    person_id = content["data"]["createPersons"]["items"][0]["id"]

    created_person = Person.objects.get(id=person_id)
    assert created_person.addresses.count() == 1

    address = created_person.addresses.first()
    assert address.street == "Teststreet"
    assert address.housenumber == "123"
    assert address.postal_code == "12345"
    assert address.place == "Testcity"
    assert address.country == "DE"


def test_edit_person_with_address(client_query, uploaded_picture):
    p = Person.objects.first()
    global_permission = Permission.objects.get(
        codename="change_person", content_type__app_label="core"
    )
    p.user.user_permissions.add(global_permission)

    p = Person.objects.create(
        first_name="Foo",
        last_name="Bar",
    )

    query = """
    mutation updatePersons($input: [BatchPatchPersonInput]!) {
      updatePersons(input: $input) {
        items: persons {
          id
          firstName
          lastName
        }
      }
    }
    """

    # Edit with adding address
    variables = {
        "input": [
            {
                "id": p.id,
                "firstName": "Foo",
                "lastName": "Bar",
                "street": "Teststreet",
                "housenumber": "123",
                "postalCode": "12345",
                "place": "Testcity",
                "country": "DE",
            },
        ]
    }
    response, content = client_query(query, variables=variables)

    p.refresh_from_db()

    assert p.addresses.count() == 1
    address = p.addresses.first()
    assert address.street == "Teststreet"
    assert address.housenumber == "123"
    assert address.postal_code == "12345"
    assert address.place == "Testcity"
    assert address.country == "DE"

    # Edit without changing address
    variables = {
        "input": [
            {
                "id": p.id,
                "firstName": "Foo",
                "lastName": "Baz",
            },
        ]
    }
    response, content = client_query(query, variables=variables)
    p.refresh_from_db()

    assert p.addresses.count() == 1
    address = p.addresses.first()
    assert address.street == "Teststreet"
    assert address.housenumber == "123"
    assert address.postal_code == "12345"
    assert address.place == "Testcity"
    assert address.country == "DE"
    # assert address.address_types.count() == 1
    # assert address.address_types.first().name == "default"

    # Edit with changing address
    variables = {
        "input": [
            {
                "id": p.id,
                "firstName": "Foo",
                "lastName": "Bar",
                "street": "Teststreetnew",
                "housenumber": "123new",
                "postalCode": "12345new",
                "place": "Testcitynew",
                "country": "GB",
            },
        ]
    }
    response, content = client_query(query, variables=variables)

    p.refresh_from_db()

    assert p.addresses.count() == 1
    address = p.addresses.first()
    assert address.street == "Teststreetnew"
    assert address.housenumber == "123new"
    assert address.postal_code == "12345new"
    assert address.place == "Testcitynew"
    assert address.country == "GB"

    # Edit with deleting address
    variables = {
        "input": [
            {
                "id": p.id,
                "firstName": "Foo",
                "lastName": "Baz",
                "street": None,
                "housenumber": None,
                "postalCode": None,
                "place": None,
                "country": None,
            },
        ]
    }
    response, content = client_query(query, variables=variables)
    p = Person.objects.get(id=p.id)

    assert p.addresses.count() == 0



import pytest

from aleksis.core.models import Person

pytestmark = pytest.mark.django_db


def test_managed_by():
    managed_person = Person.objects.create(
        first_name="Jane", last_name="Doe", managed_by_app_label="core"
    )
    unmanaged_person_1 = Person.objects.create(first_name="Jane 2", last_name="Doe 2")
    unmanaged_person_2 = Person.objects.create(first_name="Jane 2", last_name="Doe 2")

    assert list(Person.objects.managed_by_app("core")) == [managed_person]

    assert list(Person.objects.all()) == [managed_person, unmanaged_person_1, unmanaged_person_2]
    assert list(Person.objects.unmanaged()) == [unmanaged_person_1, unmanaged_person_2]



from django.core.exceptions import PermissionDenied

import graphene
from graphene_django import DjangoObjectType

from ..models import PersonalEvent
from .base import (
    BaseBatchCreateMutation,
    BaseBatchDeleteMutation,
    BaseBatchPatchMutation,
    CalendarEventBatchCreateMixin,
    CalendarEventBatchPatchMixin,
)


class PersonalEventType(DjangoObjectType):
    class Meta:
        model = PersonalEvent
        fields = (
            "id",
            "title",
            "description",
            "location",
            "datetime_start",
            "datetime_end",
            "date_start",
            "date_end",
            "owner",
            "persons",
            "groups",
        )

    timezone = graphene.String()
    recurrences = graphene.String()


class PersonalEventBatchCreateMutation(CalendarEventBatchCreateMixin, BaseBatchCreateMutation):
    class Meta:
        model = PersonalEvent
        only_fields = (
            "title",
            "description",
            "location",
            "datetime_start",
            "datetime_end",
            "timezone",
            "date_start",
            "date_end",
            "recurrences",
            "persons",
            "groups",
        )
        field_types = {
            "timezone": graphene.String(),
            "recurrences": graphene.String(),
            "location": graphene.String(),
        }
        optional_fields = ("description", "timezone", "recurrences")
        permissions = ("core.create_personal_event_rule",)

    @classmethod
    def after_create_obj(cls, root, info, data, obj, input):  # noqa
        # Overwrite after_create_obj it is used by PermissionBatchCreateMixin
        if all([len(event.persons) == 0 and len(event.groups) == 0 for event in input]):
            perms = cls._meta.permissions
        else:
            perms = ("core.create_personal_event_with_invitations_rule",)

        if not info.context.user.has_perms(perms, obj):
            raise PermissionDenied()

    @classmethod
    def before_mutate(cls, root, info, input):  # noqa
        super().before_mutate(root, info, input)
        for event in input:
            event["owner"] = info.context.user.person.id
        return input


class PersonalEventBatchDeleteMutation(BaseBatchDeleteMutation):
    class Meta:
        model = PersonalEvent
        permissions = ("core.delete_personal_event_rule",)


class PersonalEventBatchPatchMutation(CalendarEventBatchPatchMixin, BaseBatchPatchMutation):
    class Meta:
        model = PersonalEvent
        permissions = ("core.edit_personal_event_rule",)
        only_fields = (
            "id",
            "title",
            "description",
            "location",
            "datetime_start",
            "datetime_end",
            "timezone",
            "date_start",
            "date_end",
            "recurrences",
            "persons",
            "groups",
        )
        field_types = {
            "timezone": graphene.String(),
            "recurrences": graphene.String(),
            "location": graphene.String(),
        }
        optional_fields = ("description", "timezone", "recurrences")



import graphene


class HealthCheckPluginType(graphene.ObjectType):
    status = graphene.String()
    pretty_status = graphene.String()
    identifier = graphene.String()
    time_taken = graphene.Int()

    def resolve_pretty_status(root, info, **kwargs):
        return root.pretty_status()

    def resolve_identifier(root, info, **kwargs):
        return root.identifier()



from django.db import migrations, models
from django.db.models import F, Q

class Migration(migrations.Migration):

    dependencies = [
        ('core', '0070_oauth_token_checksum'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='calendarevent',
            constraint=models.CheckConstraint(condition=Q(datetime_end__gte=F('datetime_start')),
                                              name="datetime_start_before_end"
            ),
        ),
        migrations.AddConstraint(
            model_name='calendarevent',
            constraint=models.CheckConstraint(condition=Q(date_end__gte=F('date_start')),
                                              name="date_start_before_end"
            ),
        ),
    ]



import graphene
from graphene_django import DjangoObjectType

from ..models import SchoolTerm
from .base import (
    BaseBatchCreateMutation,
    BaseBatchDeleteMutation,
    BaseBatchPatchMutation,
    DjangoFilterMixin,
    PermissionsTypeMixin,
)


class SchoolTermType(PermissionsTypeMixin, DjangoFilterMixin, DjangoObjectType):
    class Meta:
        model = SchoolTerm
        filter_fields = {
            "name": ["icontains", "exact"],
            "date_start": ["exact", "lt", "lte", "gt", "gte"],
            "date_end": ["exact", "lt", "lte", "gt", "gte"],
        }
        fields = ("id", "name", "date_start", "date_end")

    current = graphene.Boolean()

    @staticmethod
    def resolve_current(root, info):
        return (current := SchoolTerm.current) and root.pk == current.pk


class SchoolTermBatchCreateMutation(BaseBatchCreateMutation):
    class Meta:
        model = SchoolTerm
        permissions = ("core.create_schoolterm_rule",)
        only_fields = ("id", "name", "date_start", "date_end")


class SchoolTermBatchDeleteMutation(BaseBatchDeleteMutation):
    class Meta:
        model = SchoolTerm
        permissions = ("core.delete_schoolterm_rule",)


class SchoolTermBatchPatchMutation(BaseBatchPatchMutation):
    class Meta:
        model = SchoolTerm
        permissions = ("core.edit_schoolterm_rule",)
        only_fields = ("id", "name", "date_start", "date_end")


class SetActiveSchoolTermMutation(graphene.Mutation):
    class Arguments:
        id = graphene.ID(required=True)  # noqa

    active_school_term = graphene.Field(SchoolTermType)

    @classmethod
    def mutate(cls, root, info, id):  # noqa
        school_term = SchoolTerm.objects.get(id=id)
        info.context.session["active_school_term"] = school_term.pk

        return SetActiveSchoolTermMutation(active_school_term=school_term)



from datetime import datetime
from xml.etree import ElementTree

from django.db.models import Q
from django.http import Http404
from django.urls import reverse

from ..core_helpers import EXTENDED_ITEM_ELEMENT_FIELD_MAP
from .base import ElementHandler, QueryBase
from .generic import DAVHref, DAVProp


class TimeRangeFilter(ElementHandler):
    _class_name = "{urn:ietf:params:xml:ns:caldav}time-range"
    invisible = True

    def post_handle(self):
        report_base = next(iter(self.request.children))

        for k, v in self.attrs.items():
            if k in [(None, "start"), (None, "end")]:
                d = datetime.fromisoformat(v)
                report_base.get_objects_args[k[1]] = d


class TextMatch(ElementHandler):
    _class_name = "{urn:ietf:params:xml:ns:caldav}text-match"
    invisible = True

    def _matches(self, obj, field_name):
        method_name = f"value_{field_name}"
        if not hasattr(obj, method_name) or not callable(getattr(obj, method_name)):
            return False

        return self.content.lower() in getattr(obj, method_name)(obj)  # FIXME: Collations

    def post_handle(self):
        field_name = self.parent._get_field_name(self.parent.attrs.get((None, "name")))
        objs = [
            obj.pk
            for obj in filter(lambda obj: self._matches(obj, field_name), self.request.objects)
        ]
        q = Q(pk__in=[objs])

        report_base = next(iter(self.request.children))

        if "additional_filter" in report_base.get_objects_args:
            q = report_base.get_objects_args["additional_filter"] & q

        report_base.get_objects_args["additional_filter"] = q


class PropFilter(ElementHandler):
    _class_name = "{urn:ietf:params:xml:ns:caldav}prop-filter"
    invisible = True

    @staticmethod
    def _get_field_name(ical_name):
        fields = list(filter(lambda f: f[1] == ical_name.lower(), EXTENDED_ITEM_ELEMENT_FIELD_MAP))
        try:
            return fields.pop()[0]
        except StopIteration:
            return None


class CalDAVProp(ElementHandler):
    _class_name = "{urn:ietf:params:xml:ns:caldav}prop"
    invisible = True

    @classmethod
    def get_name(cls, attrs):
        name = attrs.get((None, "name"))
        return f"{cls._class_name}-{name}"

    def _get_calendar_data(self, parent):
        """Helper method to find the base `CalendarData` instance."""

        if isinstance(parent, CalendarData):
            return parent
        return self._get_calendar_data(parent.parent)

    def pre_handle(self):
        calendar_data = self._get_calendar_data(self.parent.parent)
        comp_name = self.parent.attrs.get((None, "name"))
        prop = self.attrs.get((None, "name"))
        calendar_data.params.setdefault(comp_name, []).append(prop)


class CalDAVFilter(ElementHandler):
    _class_name = "{urn:ietf:params:xml:ns:caldav}filter"
    invisible = True

    def pre_handle(self):
        self.filters = {}


class CalDAVCompFilter(ElementHandler):
    _class_name = "{urn:ietf:params:xml:ns:caldav}comp-filter"
    invisible = True


class CalDAVComp(ElementHandler):
    _class_name = "{urn:ietf:params:xml:ns:caldav}comp"
    invisible = True

    @classmethod
    def get_name(cls, attrs):
        name = attrs.get((None, "name"))
        return f"{cls._class_name}-{name}"

    def pre_handle(self):
        if self.attrs.get((None, "name")) == "VTIMEZONE":
            self.parent.parent.params["VTIMEZONE"] = True


class Expand(ElementHandler):
    _class_name = "{urn:ietf:params:xml:ns:caldav}expand"

    def post_handle(self):
        for k, v in self.attrs.items():
            if k in [(None, "start"), (None, "end")]:
                d = datetime.fromisoformat(v)
                self.parent.params[f"expand_{k[1]}"] = d


class CalendarData(ElementHandler):
    _class_name = "{urn:ietf:params:xml:ns:caldav}calendar-data"

    def pre_handle(self):
        self.params = {}

    def post_handle(self):
        if not self.params:
            attrs = {(None, "name"): "VCALENDAR"}
            vcalendar = CalDAVComp(self.request, self, attrs)
            self.children.append(vcalendar)

            self.params["VTIMEZONE"] = True

            for comp_name in ("VTIMEZONE", "VEVENT"):
                attrs = {(None, "name"): comp_name}
                comp = CalDAVComp(self.request, self, attrs)
                vcalendar.children.append(comp)

    def process_xml(self, base, response):
        super().process_xml(base, response)
        if not self.invisible:
            if response.obj is not None:
                objects = response.resource.objects.filter(pk=response.obj.pk)
            else:
                objects = []

            ical = response.resource.get_dav_file_content(
                base.request._request, objects, params=self.params
            )
            base.current_xml.text = ical.decode()


class CalendarColor(ElementHandler):
    _class_name = "{http://apple.com/ns/ical/}calendar-color"

    def process_xml(self, base, response):
        if not hasattr(response.resource, "get_color"):
            raise Http404

        super().process_xml(base, response)
        base.current_xml.text = response.resource.get_color(self.request._request)


class CalendarDescription(ElementHandler):
    _class_name = "{urn:ietf:params:xml:ns:caldav}calendar-description"

    def process_xml(self, base, response):
        if not hasattr(response.resource, "get_description"):
            raise Http404

        super().process_xml(base, response)
        base.current_xml.text = response.resource.get_description()


class CalendarHomeSet(ElementHandler):
    _class_name = "{urn:ietf:params:xml:ns:caldav}calendar-home-set"

    def process_xml(self, base, response):
        super().process_xml(base, response)
        href = ElementTree.SubElement(base.current_xml, "{DAV:}href")
        href.text = reverse("dav_subregistry", args=["calendar"])


class SupportedCalendarComponentSet(ElementHandler):
    _class_name = "{urn:ietf:params:xml:ns:caldav}supported-calendar-component-set"

    def process_xml(self, base, response):
        super().process_xml(base, response)
        comp = ElementTree.SubElement(base.current_xml, "{urn:ietf:params:xml:ns:caldav}comp")
        # In AlekSIS, a calendar can only provide one type of component.
        # Therefore, the value is independent from a specific instance,
        # so that reference_object can be None.
        component_types = {
            "event": "VEVENT",
            "todo": "VTODO",
            "freebusy": "VFREEBUSY",
        }
        comp.set("name", component_types[response.resource.value_component_type(None)])


class SupportedCollationSet(ElementHandler):
    _class_name = "{urn:ietf:params:xml:ns:caldav}supported-collation-set"

    def process_xml(self, base, response):
        super().process_xml(base, response)

        supported_collations = [
            "i;ascii-casemap",
            "i;octet",
        ]

        for collation in supported_collations:
            supported_collation = ElementTree.SubElement(
                base.current_xml, "{urn:ietf:params:xml:ns:caldav}supported-collation"
            )
            supported_collation.text = collation


class Timezone(ElementHandler):
    _class_name = "{urn:ietf:params:xml:ns:caldav}timezone"
    invisible = True


class ReportBase(QueryBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.get_objects_args = {
            "request": self.request._request,
        }

    def post_handle(self):
        super().post_handle()

        try:
            timezone = next(filter(lambda child: isinstance(child, Timezone), self.children))
        except StopIteration:
            timezone = None

        if timezone is not None:
            prop = next(filter(lambda child: isinstance(child, DAVProp), self.children))
            calendar_data = next(
                filter(lambda child: isinstance(child, CalendarData), prop.children)
            )
            calendar_data.params["timezone"] = timezone.content


class CalendarQuery(ReportBase, ElementHandler):
    _class_name = "{urn:ietf:params:xml:ns:caldav}calendar-query"

    def pre_handle(self):
        self.request.resources = []

    def post_handle(self):
        self.request.objects = self.request.resource.get_objects(**self.get_objects_args)


class CalendarMultiget(ReportBase, ElementHandler):
    _class_name = "{urn:ietf:params:xml:ns:caldav}calendar-multiget"

    def pre_handle(self):
        self.request.resources = []
        self.request.objects = []

    def post_handle(self):
        super().post_handle()

        for child in self.children:
            if child._class_name == DAVHref._class_name:
                child.invisible = True



# Generated by Django 4.2.8 on 2023-12-17 18:54

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("core", "0083_availabilityevent"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="group",
            options={
                "ordering": ["short_name", "name"],
                "permissions": (
                    (
                        "assign_child_groups_to_groups",
                        "Can assign child groups to groups",
                    ),
                    ("view_group_stats", "Can view statistics about group."),
                    (
                        "view_group_free_busy_feed",
                        "Can view groups aggregated availability feed",
                    ),
                ),
                "verbose_name": "Group",
                "verbose_name_plural": "Groups",
            },
        ),
        migrations.AlterModelOptions(
            name="person",
            options={
                "ordering": ["last_name", "first_name"],
                "permissions": (
                    ("view_address", "Can view address"),
                    ("view_contact_details", "Can view contact details"),
                    ("view_photo", "Can view photo"),
                    ("view_avatar", "Can view avatar image"),
                    ("view_person_groups", "Can view persons groups"),
                    ("view_personal_details", "Can view personal details"),
                    (
                        "view_person_free_busy_feed",
                        "Can view persons aggregated availability feed",
                    ),
                ),
                "verbose_name": "Person",
                "verbose_name_plural": "Persons",
            },
        ),
    ]



from django.templatetags.static import static

import graphene


class ThemeLogoType(graphene.ObjectType):
    url = graphene.String(required=False)


class SitePreferencesType(graphene.ObjectType):
    general_title = graphene.String()

    theme_logo = graphene.Field(ThemeLogoType)
    theme_primary = graphene.String()
    theme_secondary = graphene.String()
    theme_design = graphene.String()

    footer_imprint_url = graphene.String()
    footer_privacy_url = graphene.String()

    account_person_prefer_photo = graphene.Boolean()

    calendar_days_of_the_week = graphene.List(graphene.Int)

    editable_fields_person = graphene.List(graphene.String)

    invite_enabled = graphene.Boolean()
    signup_enabled = graphene.Boolean()
    signup_required_fields = graphene.List(graphene.String)
    signup_address_required_fields = graphene.List(graphene.String)
    signup_visible_fields = graphene.List(graphene.String)
    signup_address_visible_fields = graphene.List(graphene.String)

    auth_allowed_username_regex = graphene.String()
    auth_disallowed_uids = graphene.List(graphene.String)

    def resolve_general_title(parent, info, **kwargs):
        return parent["general__title"]

    def resolve_theme_logo(parent, info, **kwargs):
        return (
            parent["theme__logo"]
            if parent["theme__logo"]
            else {"url": static("/img/aleksis-banner.svg")}
        )

    def resolve_theme_primary(parent, info, **kwargs):
        return parent["theme__primary"]

    def resolve_theme_secondary(parent, info, **kwargs):
        return parent["theme__secondary"]

    def resolve_theme_design(parent, info, **kwargs):
        if info.context.user and not info.context.user.is_anonymous and info.context.user.person:
            return info.context.user.person.preferences["theme__design"]
        return "light"

    def resolve_footer_imprint_url(parent, info, **kwargs):
        return parent["footer__imprint_url"]

    def resolve_footer_privacy_url(parent, info, **kwargs):
        return parent["footer__privacy_url"]

    def resolve_account_person_prefer_photo(parent, info, **kwargs):
        return parent["account__person_prefer_photo"]

    def resolve_calendar_days_of_the_week(parent, info, **kwargs):
        first = int(parent["calendar__first_day_of_the_week"])

        if first != 0:
            return list(range(first, 7)) + list(range(0, first))
        else:
            return list(range(first, 7))

    def resolve_editable_fields_person(parent, info, **kwargs):
        return parent["account__editable_fields_person"]

    def resolve_invite_enabled(parent, info, **kwargs):
        return parent["auth__invite_enabled"]

    def resolve_auth_allowed_username_regex(parent, info, **kwargs):
        return parent["auth__allowed_username_regex"]

    def resolve_auth_disallowed_uids(parent, info, **kwargs):
        return parent["auth__disallowed_uids"].split(",")

    def resolve_signup_enabled(parent, info, **kwargs):
        return parent["auth__signup_enabled"]

    def resolve_signup_required_fields(parent, info, **kwargs):
        return parent["auth__signup_required_fields"]

    def resolve_signup_address_required_fields(parent, info, **kwargs):
        return parent["auth__signup_address_required_fields"]

    def resolve_signup_visible_fields(parent, info, **kwargs):
        return parent["auth__signup_visible_fields"]

    def resolve_signup_address_visible_fields(parent, info, **kwargs):
        return parent["auth__signup_address_visible_fields"]



from django.core.exceptions import PermissionDenied

import graphene

from ..models import Todo


class SetTodoCompletedMutation(graphene.Mutation):
    class Arguments:
        id = graphene.ID()  # noqa
        completed = graphene.DateTime(required=False)

    ok = graphene.Boolean()

    @classmethod
    def mutate(cls, root, info, id, completed):  # noqa
        todo = Todo.objects.get(pk=id)

        if not todo.check_if_can_edit(info.context.user):
            raise PermissionDenied()

        todo.completed = completed
        if completed:
            todo.percent_complete = 100
        else:
            todo.percent_complete = 0
        todo.save()

        return SetTodoCompletedMutation(ok=True)



from django.test import override_settings
from django.urls import reverse

from aleksis.core.models import UserAdditionalAttributes

try:
    import ldap

    HAS_LDAP = True
    from django_auth_ldap.config import LDAPSearch

    LDAP_BASE = "dc=example,dc=com"
    LDAP_SETTINGS = {
        "AUTH_LDAP_GLOBAL_OPTIONS": {
            ldap.OPT_NETWORK_TIMEOUT: 1,
        },
        "AUTH_LDAP_USER_SEARCH": LDAPSearch(LDAP_BASE, ldap.SCOPE_SUBTREE),
    }
except ImportError:
    HAS_LDAP = False
    LDAP_SETTINGS = {}
import pytest

pytestmark = pytest.mark.django_db


@pytest.mark.skip(reason="broken")
@override_settings(
    AUTHENTICATION_BACKENDS=[
        "aleksis.core.util.ldap.LDAPBackend",
        "django.contrib.auth.backends.ModelBackend",
    ],
    AUTH_LDAP_SERVER_URI="ldap://[100::0]",
    AUTH_LDAP_SET_USABLE_PASSWORD=True,
    **LDAP_SETTINGS,
)
def test_login_ldap_fail_if_previously_ldap_authenticated(client, django_user_model):
    username = "foo"
    password = "bar"

    django_user_model.objects.create_user(username=username, password=password)

    # Logging in with a fresh account should success
    res = client.login(username=username, password=password)
    assert res
    client.get(reverse("logout"), follow=True)

    # Logging in with a previously LDAP-authenticated account should fail
    UserAdditionalAttributes.set_user_attribute(username, "ldap_authenticated", True)
    res = client.login(username=username, password=password)
    assert not res

    # Explicitly noting account has not been used with LDAP should succeed
    UserAdditionalAttributes.set_user_attribute(username, "ldap_authenticated", False)
    res = client.login(username=username, password=password)
    assert res


if not HAS_LDAP:
    test_login_ldap_fail_if_previously_ldap_authenticated = pytest.mark.skip(
        reason="ldap module not available"
    )(test_login_ldap_fail_if_previously_ldap_authenticated)



from datetime import date, datetime

import pytest
from recurrence import WEEKLY, Recurrence, Rule

from aleksis.core.models import Holiday

pytestmark = pytest.mark.django_db


def test_holiday_get_days():
    holiday = Holiday.objects.create(
        date_start=date(2024, 2, 1), date_end=date(2024, 2, 4), holiday_name="Test Holiday"
    )
    assert set(holiday.get_days()) == {
        date(2024, 2, 1),
        date(2024, 2, 2),
        date(2024, 2, 3),
        date(2024, 2, 4),
    }


def test_holiday_exdates():
    holiday = Holiday.objects.create(
        date_start=date(2024, 2, 1), date_end=date(2024, 2, 28), holiday_name="Test Holiday"
    )

    pattern = Recurrence(datetime(2024, 2, 3))
    pattern.rrules.append(Rule(WEEKLY, until=datetime(2024, 6, 1)))

    exdates = holiday.get_ex_dates(datetime(2024, 2, 3), datetime(2024, 4, 1), pattern)
    assert set(exdates) == {
        datetime(2024, 2, 3),
        datetime(2024, 2, 10),
        datetime(2024, 2, 17),
        datetime(2024, 2, 24),
    }



from graphene_django import DjangoObjectType

from ..models import Role
from .base import (
    BaseBatchCreateMutation,
    BaseBatchDeleteMutation,
    BaseBatchPatchMutation,
    DjangoFilterMixin,
    PermissionsTypeMixin,
)


class RoleType(PermissionsTypeMixin, DjangoFilterMixin, DjangoObjectType):
    """GraphQL type for Role"""

    class Meta:
        model = Role
        fields = [
            "id",
            "name",
            "short_name",
            "reciprocal_name",
            "ical_participation_role",
            "vcard_related_type",
            "fg_color",
            "bg_color",
        ]


class RoleBatchCreateMutation(BaseBatchCreateMutation):
    """GraphQL batch create mutation for Role"""

    class Meta:
        model = Role
        permissions = ("core.create_role_rule",)
        only_fields = (
            "name",
            "short_name",
            "reciprocal_name",
            "ical_participation_role",
            "vcard_related_type",
            "fg_color",
            "bg_color",
        )
        optional_fields = (
            "ical_participation_role",
            "vcard_related_type",
            "fg_color",
            "bg_color",
        )


class RoleBatchDeleteMutation(BaseBatchDeleteMutation):
    """GraphQL batch delete mutation for Role"""

    class Meta:
        model = Role
        permissions = ("core.delete_role_rule",)


class RoleBatchPatchMutation(BaseBatchPatchMutation):
    """GraphQL batch patch mutation for Role"""

    class Meta:
        model = Role
        permissions = ("core.edit_role_rule",)
        only_fields = (
            "id",
            "name",
            "short_name",
            "reciprocal_name",
            "ical_participation_role",
            "vcard_related_type",
            "fg_color",
            "bg_color",
        )



from django.utils.encoding import force_str

from graphene import Scalar
from graphene_django.converter import convert_django_field
from graphql import GraphQLError, StringValueNode, print_ast
from timezone_field import TimeZoneField
from timezone_field.backends import TimeZoneNotFoundError, get_tz_backend, zoneinfo


class TimeZone(Scalar):
    """
    The `TimeZone` scalar type represents a TimeZone that is accepted by django-timezone-field.

    Based on https://github.com/mfogel/django-timezone-field/blob/main/timezone_field/rest_framework.py
    and https://docs.graphene-python.org/en/latest/_modules/graphene/types/datetime/
    """

    @staticmethod
    def serialize(tz):
        return str(tz)

    @classmethod
    def parse_literal(cls, node, _variables=None):
        if not isinstance(node, StringValueNode):
            raise GraphQLError(f"TimeZone is not a string: {print_ast(node)}")
        return cls.parse_value(node.value)

    @staticmethod
    def parse_value(value):
        if isinstance(value, zoneinfo.zoneinfo.ZoneInfo):
            return value
        if not isinstance(value, str):
            raise GraphQLError(f"TimeZone is not a string: {repr(value)}")

        data_str = force_str(value)
        try:
            return get_tz_backend(use_pytz=False).to_tzobj(data_str)
        except TimeZoneNotFoundError as e:
            raise GraphQLError(f"Invalid timezone: {data_str}") from e


def setup_types():
    @convert_django_field.register(TimeZoneField)
    def convert_timezone_field(field, registry=None):
        return TimeZone(description=field.help_text, required=not field.null)



import os

from django.conf import settings
from django.core.management.base import BaseCommand

import sass


class Command(BaseCommand):
    help = "Compile SCSS stylesheet for AlekSIS"  # noqa

    def handle(self, *args, **options):
        static_dir = os.path.join(settings.BASE_DIR, "aleksis", "core", "static")
        scss_dir = os.path.join(static_dir, "public")
        scss_path = os.path.join(scss_dir, "style.scss")
        css_path = os.path.join(static_dir, "style.css")

        compiled = sass.compile(
            filename=scss_path,
            include_paths=[
                os.path.join(settings.JS_ROOT, "@materializecss", "materialize", "sass"),
                os.path.join(scss_path),
            ],
        )
        with open(css_path, "w") as f:
            f.write(compiled)



from django.core.exceptions import PermissionDenied, ValidationError
from django.utils.translation import gettext_lazy as _

import graphene
from graphene_django import DjangoObjectType

from ..data_checks import check_data
from ..models import DataCheckResult, TaskUserAssignment


class SolveOptionType(graphene.ObjectType):
    name = graphene.String()
    verbose_name = graphene.String()

    @staticmethod
    def resolve_name(root, info, **kwargs):
        return root[0]

    @staticmethod
    def resolve_verbose_name(root, info, **kwargs):
        return root[1].verbose_name


class DataCheckType(graphene.ObjectType):
    problem_name = graphene.String()
    verbose_name = graphene.String()
    solve_options = graphene.List(SolveOptionType)

    @staticmethod
    def resolve_solve_options(root, info, **kwargs):
        return root.solve_options.items()


class GenericRelatedObjectType(graphene.ObjectType):
    id = graphene.ID()
    class_verbose_name = graphene.String()
    instance_verbose_name = graphene.String()
    absolute_url = graphene.String()

    @staticmethod
    def resolve_absolute_url(root, info, **kwargs):
        return root.get_absolute_url().replace("/django", "")

    @staticmethod
    def resolve_class_verbose_name(root, info, **kwargs):
        return root._meta.verbose_name.title()

    @staticmethod
    def resolve_instance_verbose_name(root, info, **kwargs):
        return str(root)


class DataCheckResultType(DjangoObjectType):
    class Meta:
        model = DataCheckResult
        fields = (
            "id",
            "solved",
        )

    related_check = graphene.Field(DataCheckType)
    related_object = graphene.Field(GenericRelatedObjectType)


class RunDataChecksMutation(graphene.Mutation):
    task_id = graphene.ID()

    @classmethod
    def mutate(cls, root, info):
        if not info.context.user.has_perm("core.run_data_checks_rule"):
            raise PermissionDenied()

        user_assignment = check_data.delay_with_progress(
            TaskUserAssignment(
                user=info.context.user,
                title=_("Progress: Run data checks"),
                progress_title=_("Run data checks …"),
                error_message=_("There was a problem while running data checks."),
                success_message=_("The data checks were run successfully."),
                back_url="/data_checks/",
            )
        )

        return RunDataChecksMutation(task_id=user_assignment.id)


class SolveDataCheckResultMutation(graphene.Mutation):
    class Arguments:
        result = graphene.ID()
        solve_option = graphene.String()

    ok = graphene.Boolean()

    @classmethod
    def mutate(cls, root, info, result, solve_option):
        if not info.context.user.has_perm("core.solve_data_problem_rule"):
            raise PermissionDenied()

        result_obj = DataCheckResult.objects.get(pk=result)

        if solve_option in result_obj.related_check.solve_options:
            result_obj.solve(solve_option)
        else:
            raise ValidationError(_("The requested solve option does not exist"))

        return SolveDataCheckResultMutation(ok=True)



# Generated by Django 4.2.11 on 2024-05-08 08:55

from django.conf import settings
import django.contrib.postgres.fields
from django.db import migrations, models
import django.db.models.deletion

from django_pg_rrule.operations import RruleExtension
from recurrence import serialize



def _migrate_recurrences(apps, schema_editor):
    CalendarEvent = apps.get_model('core', 'CalendarEvent')
    for event in CalendarEvent.objects.using(schema_editor.connection.alias).all():
        if not event.recurrences:
            continue
        event.exdatetimes = event.recurrences.exdates
        event.rdatetimes = event.recurrences.rdates
        event.rrule = serialize(event.recurrences.rrules[0])
        if event.rrule.startswith("RRULE:"):
            event.rrule = event.rrule[6:]
        event.save()

class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('core', '0063_add_announcement_priority'),
    ]

    operations = [
        RruleExtension(),
        migrations.AddField(
            model_name='calendarevent',
            name='exdates',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.DateField(), blank=True, default=[], size=None),
        ),
        migrations.AddField(
            model_name='calendarevent',
            name='exdatetimes',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.DateTimeField(), blank=True, default=[], size=None),
        ),
        migrations.AddField(
            model_name='calendarevent',
            name='rdates',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.DateField(), blank=True, default=[], size=None),
        ),
        migrations.AddField(
            model_name='calendarevent',
            name='rdatetimes',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.DateTimeField(), blank=True, default=[], size=None),
        ),
        migrations.AddField(
            model_name='calendarevent',
            name='rrule',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.RunPython(_migrate_recurrences)
    ]



# Generated by Django 5.1.6 on 2025-03-19 11:38

import aleksis.core.mixins
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('core', '0076_remove_old_address_fields'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='OrganisationAddressThrough',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('managed_by_app_label', models.CharField(blank=True, editable=False, max_length=255, verbose_name='App label of app responsible for managing this instance')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='organisationaddressthrough',
            name='address',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.address'),
        ),
        migrations.AddField(
            model_name='organisationaddressthrough',
            name='organisation',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.organisation'),
        ),
        migrations.AddField(
            model_name='organisation',
            name='addresses',
            field=models.ManyToManyField(blank=True, related_name='organisations', through='core.OrganisationAddressThrough', to='core.address', verbose_name='Addresses'),
        ),
    ]



from django.contrib.contenttypes.models import ContentType

import graphene
from graphene_django import DjangoObjectType


class ContentTypeType(DjangoObjectType):
    class Meta:
        model = ContentType
        fields = (
            "id",
            "app_label",
            "model",
            "verbose_name",
        )

    verbose_name = graphene.String(required=True)

    @staticmethod
    def resolve_verbose_name(root, info, **kwargs):
        return root.app_labeled_name



# Generated by Django 5.1.5 on 2025-01-30 19:15

import django.core.validators
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('core', '0084_add_aggregated_availability_permissions'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.RemoveField(
            model_name='dashboardwidget',
            name='active',
        ),
        migrations.RemoveField(
            model_name='dashboardwidget',
            name='broken',
        ),
        migrations.RemoveField(
            model_name='dashboardwidget',
            name='size_l',
        ),
        migrations.RemoveField(
            model_name='dashboardwidget',
            name='size_m',
        ),
        migrations.RemoveField(
            model_name='dashboardwidget',
            name='size_s',
        ),
        migrations.RemoveField(
            model_name='dashboardwidget',
            name='size_xl',
        ),
        migrations.AddField(
            model_name='dashboardwidget',
            name='status',
            field=models.CharField(choices=[('off', 'Off'), ('ready', 'Ready'), ('broken', 'Broken'), ('on', 'On')], default='off', max_length=10, verbose_name='Status'),
        ),
        migrations.CreateModel(
            name='DashboardWidgetConfiguration',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('managed_by_app_label', models.CharField(blank=True, editable=False, max_length=255, verbose_name='App label of app responsible for managing this instance')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
                ('polymorphic_ctype', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='polymorphic_%(app_label)s.%(class)s_set+', to='contenttypes.contenttype')),
            ],
            options={
                'abstract': False,
                'base_manager_name': 'objects',
            },
        ),
        migrations.CreateModel(
            name='DashboardWidgetInstance',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('managed_by_app_label', models.CharField(blank=True, editable=False, max_length=255, verbose_name='App label of app responsible for managing this instance')),
                ('extended_data', models.JSONField(default=dict, editable=False)),
                ('x', models.PositiveSmallIntegerField(validators=[django.core.validators.MaxValueValidator(11)], verbose_name='X position')),
                ('y', models.PositiveSmallIntegerField(verbose_name='X position')),
                ('width', models.PositiveSmallIntegerField(validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(12)], verbose_name='Width')),
                ('height', models.PositiveSmallIntegerField(validators=[django.core.validators.MinValueValidator(1)], verbose_name='Height')),
                ('configuration', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.dashboardwidgetconfiguration', verbose_name='Configuration')),
                ('person', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.person', verbose_name='Person')),
                ('widget', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.dashboardwidget', verbose_name='Dashboard widget')),
            ],
            options={
                'verbose_name': 'Dashboard widget order',
                'verbose_name_plural': 'Dashboard widget orders',
            },
        ),
        migrations.DeleteModel(
            name='DashboardWidgetOrder',
        ),
    ]



import logging
from urllib.parse import urljoin

from django.conf import settings
from django.contrib.auth import get_user_model
from django.http import HttpResponseRedirect

from allauth.account.adapter import DefaultAccountAdapter
from allauth.account.utils import get_login_redirect_url
from allauth.socialaccount.adapter import DefaultSocialAccountAdapter
from invitations.adapters import BaseInvitationsAdapter

from ..templatetags.html_helpers import remove_prefix


class OurSocialAccountAdapter(DefaultSocialAccountAdapter):
    """Customised adapter that recognises other authentication mechanisms."""

    def pre_social_login(self, request, sociallogin):
        # Try to match social accounts to local accounts by their username if enabled
        if (
            settings.ALEKSIS_SOCIALACCOUNT_USERNAME_MATCHING
            and not sociallogin.is_existing
            and sociallogin.account.extra_data.get("preferred_username")
        ):
            username = sociallogin.account.extra_data["preferred_username"]
            try:
                user = get_user_model().objects.get(username=username)
                sociallogin.user = user
                logging.info(f"Match local account {username} to social account")
            except get_user_model().DoesNotExist:
                pass

    def validate_disconnect(self, account, accounts):
        """Validate whether or not the socialaccount account can be safely disconnected.

        Honours other authentication backends, i.e. ignores unusable passwords if LDAP is used.
        """
        if "django_auth_ldap.backend.LDAPBackend" in settings.AUTHENTICATION_BACKENDS:
            # Ignore upstream validation error as we do not need a usable password
            return None

        # Let upstream decide whether we can disconnect or not
        return super().validate_disconnect(account, accounts)


class OurAccountAdapter(BaseInvitationsAdapter, DefaultAccountAdapter):
    """Customised adapter to support invitations and signup."""

    def is_open_for_signup(self, request):
        # We have an own signup form, so we deactivate allauth here
        return False

    def post_login(self, request, user, *args, **kwargs):
        resp = super().post_login(request, user, *args, **kwargs)
        sociallogin = (kwargs.get("signal_kwargs") or {}).get("sociallogin")
        redirect_url = get_login_redirect_url(
            request, kwargs["redirect_url"], signup=kwargs["signup"]
        )
        redirect_url = remove_prefix(redirect_url, "/django")
        if sociallogin:
            return HttpResponseRedirect(redirect_url)
        return resp

    def get_reset_password_from_key_url(self, key):
        return urljoin(settings.BASE_URL, f"/accounts/password/reset/key/{key}")



from colorfield.fields import ColorField
from graphene_django.converter import convert_django_field, get_django_field_description

from aleksis.core.fields import WeekdayField


@convert_django_field.register(ColorField)
def color_field_converter(field, registry=None):
    from .base import Color

    return Color(description=get_django_field_description(field), required=not field.null)


@convert_django_field.register(WeekdayField)
def weekday_field_converter(field, registry=None):
    from .base import Weekday

    return Weekday(description=get_django_field_description(field), required=not field.null)



from django.contrib.auth import get_user_model
from django.contrib.auth.password_validation import validate_password
from django.core.exceptions import BadRequest, PermissionDenied, ValidationError
from django.utils.translation import gettext as _

import graphene
from allauth.account.internal import flows
from allauth.account.internal.flows.password_reset import request_password_reset
from allauth.headless.account.inputs import ResetPasswordKeyInput

from .permissions import GlobalPermissionType
from .person import PersonType


class UserType(graphene.ObjectType):
    id = graphene.ID()  # noqa
    username = graphene.String()
    first_name = graphene.String()
    last_name = graphene.String()

    is_authenticated = graphene.Boolean(required=True)
    is_anonymous = graphene.Boolean(required=True)
    is_impersonate = graphene.Boolean()

    person = graphene.Field(PersonType)

    global_permissions_by_name = graphene.List(
        GlobalPermissionType, permissions=graphene.List(graphene.String)
    )

    def resolve_global_permissions_by_name(root, info, permissions, **kwargs):
        return [
            {"name": permission_name, "result": info.context.user.has_perm(permission_name)}
            for permission_name in permissions
        ]


class UserInputType(graphene.InputObjectType):
    username = graphene.String(required=True)
    first_name = graphene.String(required=False)
    last_name = graphene.String(required=False)
    email = graphene.String(required=False)
    password = graphene.String(required=True)


class RequestPasswordResetMutation(graphene.Mutation):
    """Mutation for requesting a password reset email."""

    class Arguments:
        user_id = graphene.ID(required=False)
        email = graphene.String(required=False)

    success = graphene.Boolean()

    @classmethod
    def mutate(cls, root, info, user_id=None, email=None):
        if user_id and email:
            raise BadRequest()
        try:
            if user_id:
                user = get_user_model().objects.get(pk=user_id)
            else:
                user = get_user_model().objects.get(email=email)
        except get_user_model().DoesNotExist:
            return RequestPasswordResetMutation(success=True)

        if user_id and not info.context.user.has_perm("core.reset_user_password_rule", user):
            raise PermissionDenied()
        if email and not info.context.user.has_perm("core.reset_password_rule"):
            raise PermissionDenied()

        request_password_reset(info.context, email, [user], None)
        return RequestPasswordResetMutation(success=True)


def check_password_reset_key(root, info, key) -> bool:
    """Check if a password reset key is valid."""
    data = ResetPasswordKeyInput({"key": key})
    return data.is_valid()


class ChangePasswordMutation(graphene.Mutation):
    """Change the password of a user."""

    class Arguments:
        user_id = graphene.ID(required=False)
        old_password = graphene.String(required=False)
        reset_key = graphene.String(required=False)
        password = graphene.String(required=True)

    success = graphene.Boolean()
    user = graphene.Field(UserType)

    @classmethod
    def mutate(cls, root, info, password, user_id=None, old_password=None, reset_key=None):
        # Three valid combinations of inputs:
        # ID + PW → change password from somebody else
        # OLD_PW + PW → change password from user in request
        # KEY + PW → set password after reset for user affiliated with key

        if sum(map(bool, [user_id, old_password, reset_key])) != 1:
            raise BadRequest()

        if reset_key is not None:
            data = ResetPasswordKeyInput({"key": reset_key})
            if not data.is_valid():
                raise ValidationError("The reset key is not valid.")
            user = data.user
        elif user_id is not None:
            user = get_user_model().objects.get(pk=user_id)
            if (
                not user
                or not user.person
                or not info.context.user.has_perm("core.change_user_password_rule", user.person)
            ):
                raise PermissionDenied()
        elif old_password is not None:
            if not info.context.user.has_perm("core.change_password_rule"):
                raise ValidationError(_("Changing passwords is not allowed."))
            if not info.context.user.check_password(old_password):
                raise ValidationError(_("Please enter the correct old password."))
            user = info.context.user
        else:
            # Should be handled by the check beforehand
            raise BadRequest()

        # Validate new password
        validate_password(password, user)

        flows.password_change.change_password(user, password)

        if old_password is not None:
            flows.password_change.finalize_password_change(info.context, user)
        else:
            flows.password_change.finalize_password_set(info.context, user)

        return ChangePasswordMutation(success=True, user=user)



import graphene
from graphene_django import DjangoObjectType
from guardian.shortcuts import get_objects_for_user

from ..models import Group, Person, PersonGroupThrough
from ..util.core_helpers import has_person
from .base import BaseBatchDeleteMutation, BaseObjectType, DjangoFilterMixin, PermissionsTypeMixin


class GroupStatisticsType(graphene.ObjectType):
    members = graphene.Int()
    age_avg = graphene.Float()
    age_range_min = graphene.Int()
    age_range_max = graphene.Int()


class PersonGroupThroughType(BaseObjectType):
    """GraphQL type for PersonGroupThrough"""

    class Meta:
        model = PersonGroupThrough
        fields = [
            "group",
            "person",
            "roles",
        ]


class GroupType(PermissionsTypeMixin, DjangoFilterMixin, DjangoObjectType):
    class Meta:
        model = Group
        fields = [
            "id",
            "school_term",
            "name",
            "short_name",
            "members",
            "owners",
            "child_groups",
            "parent_groups",
            "group_type",
            "photo",
            "avatar",
        ]
        filter_fields = {
            "name": ["icontains"],
            "short_name": ["icontains"],
            "group_type": ["exact", "in"],
        }

    avatar_url = graphene.String()
    statistics = graphene.Field(GroupStatisticsType)

    relationships = graphene.List(PersonGroupThroughType)

    @staticmethod
    def resolve_parent_groups(root, info, **kwargs):
        qs = root.parent_groups.all()
        if info.context.user.has_perm("core.view_group_rule", root):
            return qs
        return get_objects_for_user(info.context.user, "core.view_group", qs)

    @staticmethod
    def resolve_child_groups(root, info, **kwargs):
        qs = root.child_groups.all()
        if info.context.user.has_perm("core.view_group_rule", root):
            return qs
        return get_objects_for_user(info.context.user, "core.view_group", qs)

    @staticmethod
    def resolve_members(root, info, **kwargs):
        if (
            has_person(info.context.user)
            and root.group_type
            and root.group_type.owners_can_see_members
            and info.context.user.person in root.owners.all()
        ):
            return root.members.all()
        persons = get_objects_for_user(info.context.user, "core.view_person", root.members.all())
        if has_person(info.context.user) and [
            m for m in root.members.all() if m.pk == info.context.user.person.pk
        ]:
            persons = (persons | Person.objects.filter(pk=info.context.user.person.pk)).distinct()
        return persons

    @staticmethod
    def resolve_relationships(root: Group, info, **kwargs):
        persons = GroupType.resolve_members(root, info, **kwargs)
        return PersonGroupThrough.objects.filter(group=root, person__in=persons)

    @staticmethod
    def resolve_owners(root, info, **kwargs):
        persons = get_objects_for_user(info.context.user, "core.view_person", root.owners.all())
        if has_person(info.context.user) and [
            o for o in root.owners.all() if o.pk == info.context.user.person.pk
        ]:
            persons = (persons | Person.objects.filter(pk=info.context.user.person.pk)).distinct()
        return persons

    @staticmethod
    def resolve_statistics(root: Group, info, **kwargs):
        if not info.context.user.has_perm("core.view_group_stats_rule", root):
            return None
        return root.get_group_stats

    @staticmethod
    def resolve_can_edit(root, info, **kwargs):
        if hasattr(root, "can_edit"):
            return root.can_edit
        return info.context.user.has_perm("core.edit_group_rule", root)

    @staticmethod
    def resolve_can_delete(root, info, **kwargs):
        if hasattr(root, "can_delete"):
            return root.can_delete
        return info.context.user.has_perm("core.delete_group_rule", root)


class GroupBatchDeleteMutation(BaseBatchDeleteMutation):
    class Meta:
        model = Group
        permissions = ("core.delete_group_rule",)



# Generated by Django 5.2.6 on 2025-09-25 13:22

import aleksis.core.mixins
import ckeditor.fields
import colorfield.fields
import django.contrib.postgres.fields
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0093_announcement_global_alarm'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='persongroupthrough',
            name='date_end',
            field=models.DateField(blank=True, null=True, verbose_name='End date'),
        ),
        migrations.AddField(
            model_name='persongroupthrough',
            name='date_start',
            field=models.DateField(blank=True, null=True, verbose_name='Start date'),
        ),
        migrations.AddField(
            model_name='personrelationship',
            name='date_end',
            field=models.DateField(blank=True, null=True, verbose_name='End date'),
        ),
        migrations.AddField(
            model_name='personrelationship',
            name='date_start',
            field=models.DateField(blank=True, null=True, verbose_name='Start date'),
        ),
    ]



# Generated by Django 4.2.11 on 2024-05-14 14:18

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0062_oauth_update"),
    ]

    operations = [
        migrations.AddField(
            model_name="announcement",
            name="priority",
            field=models.PositiveSmallIntegerField(blank=True, null=True, verbose_name="Priority"),
        ),
    ]



# Generated by Django 5.2.4 on 2025-07-14 20:38

import aleksis.core.mixins
import ckeditor.fields
import colorfield.fields
import django.contrib.postgres.fields
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0092_add_todo'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='announcement',
            name='is_global',
            field=models.BooleanField(default=False, help_text='Show this announcement for everyone', verbose_name='Global announcement'),
        ),
        migrations.CreateModel(
            name='AnnouncementAlarm',
            fields=[
                ('calendaralarm_ptr',
                 models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True,
                                      primary_key=True, serialize=False, to='core.calendaralarm')),
            ],
            options={
                'verbose_name': 'Announcement alarm',
                'verbose_name_plural': 'Announcement alarms',
            },
            bases=('core.calendaralarm',),
        ),
    ]



from textwrap import wrap
from typing import Any, ClassVar, Optional
from urllib.parse import urlencode, urlparse, urlunparse
from uuid import UUID
from xml.etree import ElementTree
from xml.sax import SAXParseException

from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import Group as DjangoGroup
from django.contrib.auth.models import Permission, User
from django.contrib.auth.views import LogoutView
from django.core.exceptions import BadRequest, ObjectDoesNotExist, PermissionDenied, ValidationError
from django.http import (
    Http404,
    HttpRequest,
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseRedirect,
    HttpResponseServerError,
    JsonResponse,
    QueryDict,
)
from django.shortcuts import get_object_or_404, redirect, render
from django.template import loader
from django.templatetags.static import static
from django.urls import reverse, reverse_lazy
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.utils.translation import get_language
from django.utils.translation import gettext_lazy as _
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_exempt
from django.views.defaults import ERROR_500_TEMPLATE_NAME
from django.views.generic.base import TemplateView, View
from django.views.generic.detail import DetailView, SingleObjectMixin
from django.views.generic.edit import DeleteView, FormView

import rest_framework
import reversion
from allauth.account.views import LoginView as AllAuthLoginView
from allauth.socialaccount.adapter import get_adapter
from allauth.socialaccount.models import SocialAccount
from django_filters.views import FilterView
from django_tables2 import SingleTableMixin, SingleTableView
from dynamic_preferences.forms import preference_form_builder
from graphene.validation import depth_limit_validator
from graphene_file_upload.django import FileUploadGraphQLView
from graphql import (
    ExecutionResult,
    GraphQLError,
    parse,
    validate,
)
from guardian.shortcuts import GroupObjectPermission, UserObjectPermission
from haystack.generic_views import SearchView
from haystack.inputs import AutoQuery
from haystack.query import SearchQuerySet
from haystack.utils.loading import UnifiedIndex
from health_check.views import MainView
from invitations.views import SendInvite
from oauth2_provider.contrib.rest_framework import OAuth2Authentication
from oauth2_provider.exceptions import OAuthToolkitError
from oauth2_provider.models import get_application_model
from oauth2_provider.views import AuthorizationView
from rest_framework.renderers import JSONRenderer, TemplateHTMLRenderer
from rest_framework.request import Request as APIRequest
from rest_framework.response import Response as APIResponse
from rest_framework.views import APIView
from reversion import set_user
from rules.contrib.views import PermissionRequiredMixin, permission_required

from .decorators import pwa_cache
from .filters import (
    GroupGlobalPermissionFilter,
    GroupObjectPermissionFilter,
    UserGlobalPermissionFilter,
    UserObjectPermissionFilter,
)
from .forms import (
    AssignPermissionForm,
    EditGroupForm,
    GroupPreferenceForm,
    PersonPreferenceForm,
    SelectPermissionForm,
    SitePreferenceForm,
)
from .mixins import (
    AdvancedDeleteView,
    CalendarEventMixin,
    DAVResource,
    ExtensibleModel,
    ObjectAuthenticator,
    RegistryObject,
    SuccessNextMixin,
)
from .models import (
    Group,
    Person,
    PersonInvitation,
)
from .registries import (
    group_preferences_registry,
    person_preferences_registry,
    site_preferences_registry,
)
from .schema import schema
from .tables import (
    GroupGlobalPermissionTable,
    GroupObjectPermissionTable,
    InvitationsTable,
    UserGlobalPermissionTable,
    UserObjectPermissionTable,
)
from .util import messages
from .util.auth_helpers import BasicAuthMixin
from .util.core_helpers import (
    generate_random_code,
    get_allowed_object_ids,
    get_pwa_icons,
    get_site_preferences,
    has_person,
    objectgetter_optional,
)
from .util.dav_handler.base import DAVMultistatus, DAVRequest
from .util.forms import PreferenceLayout
from .util.pdf import render_pdf

if settings.SENTRY_ENABLED:
    import sentry_sdk


class LogoView(View):
    def get(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        image = get_site_preferences()["theme__logo"]
        image_url = image.url if image else static("img/aleksis-banner.svg")

        return redirect(image_url)


class RenderPDFView(TemplateView):
    """View to render a PDF file from a template.

    Makes use of ``render_pdf``.
    """

    def get(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        context = self.get_context_data(**kwargs)
        return render_pdf(request, self.template_name, context)


class ServiceWorkerView(View):
    """Render serviceworker.js under root URL.

    This can't be done by static files,
    because the PWA has a scope and
    only accepts service worker files from the root URL.
    """

    def get(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        with open(settings.SERVICE_WORKER_PATH) as f:
            return HttpResponse(f, content_type="application/javascript")


class ManifestView(View):
    """Build manifest.json for PWA."""

    def get(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        prefs = get_site_preferences()
        pwa_imgs = get_pwa_icons()

        icons = [
            {
                "src": favicon_img.faviconImage.url,
                "sizes": f"{favicon_img.size}x{favicon_img.size}",
                "purpose": "any maskable" if prefs["theme__pwa_icon_maskable"] else "any",
            }
            for favicon_img in pwa_imgs
        ]

        manifest = {
            "name": prefs["general__title"],
            "short_name": prefs["general__title"],
            "description": prefs["general__description"],
            "start_url": "/",
            "scope": "/",
            "lang": get_language(),
            "display": "standalone",
            "orientation": "any",
            "status_bar": "default",
            "background_color": "#ffffff",
            "theme_color": prefs["theme__primary"],
            "icons": icons,
        }
        return JsonResponse(manifest)


@method_decorator(pwa_cache, name="dispatch")
class OfflineView(TemplateView):
    """Show an error page if there is no internet connection."""

    template_name = "offline.html"


def get_group_by_id(request: HttpRequest, id_: Optional[int] = None):
    if id_:
        return get_object_or_404(Group, id=id_)
    else:
        return None


@never_cache
@permission_required("core.edit_group_rule", fn=objectgetter_optional(Group, None, False))
def edit_group(request: HttpRequest, id_: Optional[int] = None) -> HttpResponse:
    """View to edit or create a group."""
    context = {}

    group = objectgetter_optional(Group, None, False)(request, id_)
    context["group"] = group

    if id_:
        # Edit form for existing group
        edit_group_form = EditGroupForm(request.POST or None, request.FILES or None, instance=group)
    else:
        # Empty form to create a new group
        if request.user.has_perm("core.create_group_rule"):
            edit_group_form = EditGroupForm(request.POST or None, request.FILES or None)
        else:
            raise PermissionDenied()

    if edit_group_form.is_valid():
        with reversion.create_revision():
            set_user(request.user)
            group = edit_group_form.save(commit=True)

        messages.success(request, _("The group has been saved."))

        return redirect("group_by_id", group.pk)

    context["edit_group_form"] = edit_group_form

    return render(request, "core/group/edit.html", context)


class SystemStatusAPIView(PermissionRequiredMixin, MainView):
    """Provide information about system status as JSON."""

    permission_required = "core.view_system_status_rule"

    @method_decorator(never_cache)
    def get(self, request, *args, **kwargs):
        status_code = 500 if self.errors else 200

        return self.render_to_response_json(self.plugins, status_code)


class TestPDFGenerationView(PermissionRequiredMixin, RenderPDFView):
    template_name = "core/pages/test_pdf.html"
    permission_required = "core.test_pdf_rule"


@permission_required("core.search_rule")
def searchbar_snippets(request: HttpRequest) -> HttpResponse:
    """View to return HTML snippet with searchbar autocompletion results."""
    query = request.GET.get("q", "")
    limit = int(request.GET.get("limit", "5"))
    indexed_models = UnifiedIndex().get_indexed_models()
    allowed_object_ids = get_allowed_object_ids(request, indexed_models)
    results = (
        SearchQuerySet().filter(id__in=allowed_object_ids).filter(text=AutoQuery(query))[:limit]
    )
    context = {"results": results}

    return render(request, "search/searchbar_snippets.html", context)


@method_decorator(pwa_cache, name="dispatch")
class PermissionSearchView(PermissionRequiredMixin, SearchView):
    """Wrapper to apply permission to haystack's search view."""

    permission_required = "core.search_rule"

    def get_context_data(self, *, object_list=None, **kwargs):
        queryset = object_list if object_list is not None else self.object_list
        indexed_models = UnifiedIndex().get_indexed_models()
        allowed_object_ids = get_allowed_object_ids(self.request, indexed_models)
        queryset = queryset.filter(id__in=allowed_object_ids)

        return super().get_context_data(object_list=queryset, **kwargs)


@never_cache
def preferences(
    request: HttpRequest,
    registry_name: str = "person",
    pk: Optional[int] = None,
    section: Optional[str] = None,
) -> HttpResponse:
    """View for changing preferences."""
    context = {}

    # Decide which registry to use and check preferences
    if registry_name == "site":
        registry = site_preferences_registry
        instance = None
        form_class = SitePreferenceForm

        if not request.user.has_perm("core.change_site_preferences_rule", instance):
            raise PermissionDenied()
    elif registry_name == "person":
        registry = person_preferences_registry
        instance = objectgetter_optional(Person, "request.user.person", True)(request, pk)
        form_class = PersonPreferenceForm

        if not request.user.has_perm("core.change_person_preferences_rule", instance):
            raise PermissionDenied()
    elif registry_name == "group":
        registry = group_preferences_registry
        instance = objectgetter_optional(Group, None, False)(request, pk)
        form_class = GroupPreferenceForm

        if not request.user.has_perm("core.change_group_preferences_rule", instance):
            raise PermissionDenied()
    else:
        # Invalid registry name passed from URL
        raise Http404(_("The requested preference registry does not exist"))

    if not section and len(registry.sections()) > 0:
        default_section = list(registry.sections())[0]
        if instance:
            return redirect(f"preferences_{registry_name}", instance.pk, default_section)
        else:
            return redirect(f"preferences_{registry_name}", default_section)

    # Build final form from dynamic-preferences
    form_class = preference_form_builder(form_class, instance=instance, section=section)

    # Get layout
    form_class.layout = PreferenceLayout(form_class, section=section)

    if request.method == "POST":
        form = form_class(request.POST, request.FILES or None)
        if form.is_valid():
            form.update_preferences()
            messages.success(request, _("The preferences have been saved successfully."))
    else:
        form = form_class()

    context["registry"] = registry
    context["registry_name"] = registry_name
    context["section"] = section
    context["registry_url"] = "preferences_" + registry_name
    context["form"] = form
    context["instance"] = instance

    return render(request, "dynamic_preferences/form.html", context)


@permission_required("core.delete_group_rule", fn=objectgetter_optional(Group))
def delete_group(request: HttpRequest, id_: int) -> HttpResponse:
    """View to delete an group."""
    group = objectgetter_optional(Group)(request, id_)
    with reversion.create_revision():
        set_user(request.user)
        group.save()

    group.delete()
    messages.success(request, _("The group has been deleted."))

    return redirect("groups")


class InvitePerson(PermissionRequiredMixin, SingleTableView, SendInvite):
    """View to invite a person to register an account."""

    template_name = "invitations/forms/_invite.html"
    permission_required = "core.invite_rule"
    model = PersonInvitation
    table_class = InvitationsTable
    context = {}

    def dispatch(self, request, *args, **kwargs):
        if not get_site_preferences()["auth__invite_enabled"]:
            return HttpResponseRedirect(reverse_lazy("invite_disabled"))
        return super().dispatch(request, *args, **kwargs)

    # Get queryset of invitations
    def get_context_data(self, **kwargs):
        queryset = kwargs.pop("object_list", None)
        if queryset is None:
            self.object_list = self.model.objects.all()
        return super().get_context_data(**kwargs)


class GenerateInvitationCode(View):
    """View to generate an invitation code."""

    def get(self, request):
        # Build code
        length = get_site_preferences()["auth__invite_code_length"]
        packet_size = get_site_preferences()["auth__invite_code_packet_size"]
        code = generate_random_code(length, packet_size)

        # Create invitation object
        invitation = PersonInvitation.objects.create(
            email="", inviter=request.user, key=code, sent=timezone.now()
        )

        # Make code more readable
        code = "-".join(wrap(invitation.key, 5))

        # Generate success message and print code
        messages.success(
            request,
            _(f"The invitation was successfully created. The invitation code is {code}"),
        )

        return redirect("invite_person")


@method_decorator(pwa_cache, name="dispatch")
class PermissionsListBaseView(PermissionRequiredMixin, SingleTableMixin, FilterView):
    """Base view for list of all permissions."""

    template_name = "core/perms/list.html"
    permission_required = "core.manage_permissions_rule"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["assign_form"] = SelectPermissionForm()
        context["tab"] = self.tab

        return context


@method_decorator(pwa_cache, name="dispatch")
class UserGlobalPermissionsListBaseView(PermissionsListBaseView):
    """List all global user permissions."""

    filterset_class = UserGlobalPermissionFilter
    table_class = UserGlobalPermissionTable
    tab = "user_global"


@method_decorator(pwa_cache, name="dispatch")
class GroupGlobalPermissionsListBaseView(PermissionsListBaseView):
    """List all global group permissions."""

    filterset_class = GroupGlobalPermissionFilter
    table_class = GroupGlobalPermissionTable
    tab = "group_global"


@method_decorator(pwa_cache, name="dispatch")
class UserObjectPermissionsListBaseView(PermissionsListBaseView):
    """List all object user permissions."""

    filterset_class = UserObjectPermissionFilter
    table_class = UserObjectPermissionTable
    tab = "user_object"


@method_decorator(pwa_cache, name="dispatch")
class GroupObjectPermissionsListBaseView(PermissionsListBaseView):
    """List all object group permissions."""

    filterset_class = GroupObjectPermissionFilter
    table_class = GroupObjectPermissionTable
    tab = "group_object"


@method_decorator(pwa_cache, name="dispatch")
class SelectPermissionForAssignView(PermissionRequiredMixin, FormView):
    """View for selecting a permission to assign."""

    permission_required = "core.manage_permissions_rule"
    form_class = SelectPermissionForm

    def form_valid(self, form: SelectPermissionForm) -> HttpResponse:
        url = reverse("assign_permission", args=[form.cleaned_data["selected_permission"].pk])
        params = {"next": self.request.GET["next"]} if "next" in self.request.GET else {}
        return redirect(f"{url}?{urlencode(params)}")

    def form_invalid(self, form: SelectPermissionForm) -> HttpResponse:
        return redirect("manage_group_object_permissions")


class AssignPermissionView(SuccessNextMixin, PermissionRequiredMixin, DetailView, FormView):
    """View for assigning a permission to users/groups for all/some objects."""

    permission_required = "core.manage_permissions"
    queryset = Permission.objects.all()
    template_name = "core/perms/assign.html"
    form_class = AssignPermissionForm
    success_url = "manage_user_global_permissions"

    def get_form_kwargs(self) -> dict[str, Any]:
        kwargs = super().get_form_kwargs()
        kwargs["permission"] = self.get_object()
        return kwargs

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        # Overwrite get_context_data to ensure correct function call order
        self.object = self.get_object()
        context = super().get_context_data(**kwargs)
        return context

    def form_valid(self, form: AssignPermissionForm) -> HttpResponse:
        form.save_perms()
        messages.success(
            self.request,
            _("We have successfully assigned the permissions."),
        )
        return redirect(self.get_success_url())


class UserGlobalPermissionDeleteView(PermissionRequiredMixin, AdvancedDeleteView):
    """Delete a global user permission."""

    permission_required = "core.manage_permissions"
    model = User.user_permissions.through
    success_message = _("The global user permission has been deleted.")
    success_url = reverse_lazy("manage_user_global_permissions")
    template_name = "core/pages/delete.html"


class GroupGlobalPermissionDeleteView(PermissionRequiredMixin, AdvancedDeleteView):
    """Delete a global group permission."""

    permission_required = "core.manage_permissions"
    model = DjangoGroup.permissions.through
    success_message = _("The global group permission has been deleted.")
    success_url = reverse_lazy("manage_group_global_permissions")
    template_name = "core/pages/delete.html"


class UserObjectPermissionDeleteView(PermissionRequiredMixin, AdvancedDeleteView):
    """Delete a object user permission."""

    permission_required = "core.manage_permissions"
    model = UserObjectPermission
    success_message = _("The object user permission has been deleted.")
    success_url = reverse_lazy("manage_user_object_permissions")
    template_name = "core/pages/delete.html"


class GroupObjectPermissionDeleteView(PermissionRequiredMixin, AdvancedDeleteView):
    """Delete a object group permission."""

    permission_required = "core.manage_permissions"
    model = GroupObjectPermission
    success_message = _("The object group permission has been deleted.")
    success_url = reverse_lazy("manage_group_object_permissions")
    template_name = "core/pages/delete.html"


class SocialAccountDeleteView(DeleteView):
    """Custom view to delete django-allauth social account."""

    template_name = "core/pages/delete.html"
    success_url = reverse_lazy("socialaccount_connections")

    def get_queryset(self):
        return SocialAccount.objects.filter(user=self.request.user)

    def form_valid(self, form):
        self.object = self.get_object()
        try:
            get_adapter(self.request).validate_disconnect(
                self.object, SocialAccount.objects.filter(user=self.request.user)
            )
        except ValidationError:
            messages.error(
                self.request,
                _(
                    "The third-party account could not be disconnected "
                    "because it is the only login method available."
                ),
            )
        else:
            self.object.delete()
            messages.success(
                self.request, _("The third-party account has been successfully disconnected.")
            )
        return super().form_valid()


def server_error(
    request: HttpRequest, template_name: str = ERROR_500_TEMPLATE_NAME
) -> HttpResponseServerError:
    """Ensure the request is passed to the error page."""
    template = loader.get_template(template_name)
    context = {"request": request}

    return HttpResponseServerError(template.render(context))


class InvitePersonByID(PermissionRequiredMixin, SingleObjectMixin, View):
    """Custom view to invite person by their ID."""

    model = Person
    success_url = reverse_lazy("persons")
    permission_required = "core.invite_rule"

    def dispatch(self, request, *args, **kwargs):
        if not get_site_preferences()["auth__invite_enabled"]:
            return HttpResponseRedirect(reverse_lazy("invite_disabled"))
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        person = self.get_object()

        if not person.email or not PersonInvitation.objects.filter(email=person.email).exists():
            length = get_site_preferences()["auth__invite_code_length"]
            packet_size = get_site_preferences()["auth__invite_code_packet_size"]
            key = generate_random_code(length, packet_size)
            invite = PersonInvitation.objects.create(person=person, key=key)
            if person.email:
                invite.email = person.email
            invite.inviter = self.request.user
            invite.save()

            invite.send_invitation(self.request)

            if person.email:
                messages.success(
                    self.request,
                    _(
                        "Person was invited successfully and an email "
                        "with further instructions has been send to them."
                    ),
                )
            else:
                readable_key = "-".join(wrap(key, packet_size))
                messages.success(
                    self.request,
                    f"{_('Person was invited successfully. Their key is')} {readable_key}.",
                )
        else:
            messages.success(self.request, _("Person was already invited."))

        return redirect("person_by_id", person.pk)


class InviteDisabledView(PermissionRequiredMixin, TemplateView):
    """View to display a notice that the invite feature is disabled and how to enable it."""

    template_name = "invitations/disabled.html"
    permission_required = "core.change_site_preferences_rule"

    def dispatch(self, request, *args, **kwargs):
        if get_site_preferences()["auth__invite_enabled"]:
            raise PermissionDenied()
        return super().dispatch(request, *args, **kwargs)


class LoginView(AllAuthLoginView):
    """Custom login view."""

    def get_context_data(self, **kwargs):
        """Override context data to hide side menu and include OAuth2 application if given."""
        context = super().get_context_data(**kwargs)
        if self.request.GET.get("oauth") and self.request.GET.get("client_id"):
            application = get_application_model().objects.get(
                client_id=self.request.GET["client_id"]
            )
            context["oauth_application"] = application
        return context


class CustomAuthorizationView(AuthorizationView):
    def handle_no_permission(self):
        """Override handle_no_permission to provide OAuth2 information to login page."""
        redirect_obj = super().handle_no_permission()

        try:
            scopes, credentials = self.validate_authorization_request(self.request)
        except OAuthToolkitError as error:
            # Application is not available at this time.
            return self.error_response(error, application=None)

        login_url_parts = list(urlparse(redirect_obj.url))
        querystring = QueryDict(login_url_parts[4], mutable=True)
        querystring["oauth"] = "yes"
        querystring["client_id"] = credentials["client_id"]
        login_url_parts[4] = querystring.urlencode(safe="/")

        return HttpResponseRedirect(urlunparse(login_url_parts))

    def get_context_data(self, **kwargs):
        """Override context data to hide side menu."""
        context = super().get_context_data(**kwargs)
        context["no_menu"] = True
        return context


class LoggingGraphQLView(FileUploadGraphQLView):
    """GraphQL view that raises unknown exceptions instead of blindly catching them."""

    def execute_graphql_request(
        self, request, data, query, variables, operation_name, show_graphiql=False
    ):
        if settings.SENTRY_ENABLED and operation_name:
            scope = sentry_sdk.get_current_scope()
            scope.set_transaction_name(operation_name)

        validation_errors = []

        if query:
            validation_errors = validate(
                schema=schema.graphql_schema,
                document_ast=parse(query),
                rules=(depth_limit_validator(max_depth=10),),
            )
        if validation_errors:
            result = ExecutionResult(data=None, errors=validation_errors)
        else:
            result = super().execute_graphql_request(
                request, data, query, variables, operation_name, show_graphiql
            )

        errors = result.errors or []
        for error in errors:
            if not isinstance(
                error.original_error,
                (GraphQLError, ValidationError, PermissionDenied, ObjectDoesNotExist),
            ):
                if isinstance(error.original_error, ObjectDoesNotExist):
                    raise GraphQLError from PermissionDenied
                raise error

            # Pass validation errors to frontend
            if isinstance(error.original_error, ValidationError):
                error.message = (
                    error.original_error.message_dict
                    if getattr(error.original_error, "error_dict", None)
                    else error.original_error.messages
                )

        return result


class ObjectAuthentication(rest_framework.authentication.BaseAuthentication):
    def authenticate(self, request: APIRequest):
        authenticators = request.GET.get("authenticators", "").split(",")
        if authenticators == [""]:
            authenticators = list(ObjectAuthenticator.registered_objects_dict.keys())

        for authenticator in authenticators:
            authenticator_class = ObjectAuthenticator.get_object_by_name(authenticator)
            if not authenticator_class:
                continue
            res = authenticator_class().authenticate(request, request._aleksis_object)
            if res is None:
                return None
            user, obj = res
            if request._aleksis_object is not None and obj != request._aleksis_object:
                raise BadRequest("Ambiguous objects identified")
            if request._aleksis_object is None:
                request._aleksis_object = obj
            if user is None:
                return None
            return user, obj

        return None


class ObjectSerializer(rest_framework.serializers.BaseSerializer):
    def to_representation(self, instance):
        if hasattr(instance, "get_json"):
            res = instance.get_json(self._context["request"])
        else:
            res = {"id": instance.id}
        res["_meta"] = {
            "model": instance._meta.model_name,
            "app": instance._meta.app_label,
        }

        return res


class ObjectRepresentationView(APIView):
    """View with unique URL to get a JSON representation of an object."""

    authentication_classes = [
        ObjectAuthentication,
        rest_framework.authentication.BasicAuthentication,
        OAuth2Authentication,
        rest_framework.authentication.SessionAuthentication,
    ]
    permission_classes = []
    renderer_classes = [TemplateHTMLRenderer, JSONRenderer]

    def get_model(self, request: APIRequest, app_label: str, model: str):
        """Get the model by app label and model name."""
        try:
            model = apps.get_model(app_label, model)
            if ExtensibleModel in model.__mro__:
                return model
            else:
                raise Http404
        except LookupError as exc:
            raise Http404 from exc

    def get_object(self, request: APIRequest, app_label: str, model: str, pk: int | UUID):
        """Get the object by app label, model name and primary key."""
        if getattr(self, "model", None) is None:
            self.model = self.get_model(request, app_label, model)

        try:
            if isinstance(pk, UUID):
                return self.model.objects.get(uuid=pk)
            else:
                return self.model.objects.get(pk=pk)
        except self.model.DoesNotExist as exc:
            raise Http404 from exc

    def initial(
        self,
        request: APIRequest,
        app_label: Optional[str] = None,
        model: Optional[str] = None,
        pk: Optional[int | UUID] = None,
        *args,
        **kwargs,
    ):
        if app_label and model:
            request._aleksis_model = self.get_model(request, app_label, model)
        else:
            request._aleksis_model = None

        if app_label and model and pk:
            request._aleksis_object = self.get_object(request, app_label, model, pk)
        else:
            request._aleksis_object = None

        request._aleksis_orig_object = request._aleksis_object
        return super().initial(request, *args, **kwargs)

    def get(
        self,
        request: APIRequest,
        *args,
        **kwargs,
    ) -> HttpResponse:
        headers = {}
        status = 200
        template_name = None

        if request._aleksis_object is None:
            if has_person(request.user):
                target = request.user.person.get_object_uri()
                if qs := request.GET.urlencode():
                    target += f"?{qs}"
                return HttpResponseRedirect(target)
        elif request._aleksis_object != request._aleksis_orig_object:
            target = request._aleksis_object.get_object_uri()
            if qs := request.GET.urlencode():
                target += f"?{qs}"
            return HttpResponseRedirect(target)

        if isinstance(request.accepted_renderer, TemplateHTMLRenderer):
            if settings.OBJECT_REPR_USE_FRONTEND:
                return APIResponse(None, template_name="core/vue_index.html")
            elif request._aleksis_object is None:
                raise Http404
            else:
                if hasattr(request._aleksis_object, "get_template"):
                    template_name = request._aleksis_object.get_template()
                else:
                    template_name = "core/406.html"
                    status = 406

        if request.user.is_authenticated:
            serializer = ObjectSerializer(request._aleksis_object, context={"request": request})
            return APIResponse(
                serializer.data, status=status, headers=headers, template_name=template_name
            )

        raise PermissionDenied()


class RegistryObjectViewMixin:
    """Provide single registry object by its name."""

    registry_object: ClassVar[type[RegistryObject] | RegistryObject] = None

    def get_object(self) -> type[RegistryObject]:
        if not self.registry_object:
            raise NotImplementedError("There is no registry object set.")

        if "name" in self.kwargs and "subregistry" not in self.kwargs:
            return self._get_sub_registry()

        name = self.kwargs.get("name")
        if name is None:
            return self.registry_object
        if name in self.registry_object.registered_objects_dict:
            return self.registry_object.registered_objects_dict[name]
        raise Http404

    def _get_sub_registry(self) -> type[RegistryObject]:
        name = self.kwargs.get("name")
        if name in self.registry_object.get_sub_registries():
            return self.registry_object.get_sub_registry_by_name(name)
        raise Http404


class ICalFeedView(RegistryObjectViewMixin, PermissionRequiredMixin, View):
    """View to generate an iCal feed for a calendar."""

    permission_required = "core.view_calendar_feed_rule"
    registry_object = CalendarEventMixin

    def get(self, request, name, *args, **kwargs):
        calendar = self.get_object()
        feed = calendar.create_feed(request)
        response = HttpResponse(content_type="text/calendar")
        feed.write(response)
        return response


class ICalAllFeedsView(PermissionRequiredMixin, View):
    """View to generate an iCal feed for all calendars."""

    permission_required = "core.view_calendar_feed_rule"

    def get(self, request, *args, **kwargs):
        response = HttpResponse(content_type="text/calendar")
        for calendar in CalendarEventMixin.get_enabled_feeds(request):
            feed = calendar.create_feed(request)
            feed.write(response)
        return response


@method_decorator(csrf_exempt, name="dispatch")
class DAVResourceView(BasicAuthMixin, PermissionRequiredMixin, RegistryObjectViewMixin, View):
    """View for CalDAV collections."""

    registry_object = DAVResource
    permission_required = "core.view_calendar_feed_rule"

    http_method_names = View.http_method_names + ["propfind", "report"]
    dav_compliance = []

    _dav_request: ClassVar[DAVRequest]

    def dispatch(self, request, *args, **kwargs):
        res = super().dispatch(request, *args, **kwargs)

        res.headers["DAV"] = ", ".join(
            ["1", "3", "calendar-access", "addressbook"] + self.dav_compliance
        )

        return res

    def options(self, request, *args, **kwargs):
        res = super().options(request, *args, **kwargs)
        return res

    def propfind(self, request, *args, **kwargs):
        resource = self.get_object()
        self._dav_request = DAVRequest(request, resource, None)

        try:
            self._dav_request.parse()
        except SAXParseException:
            return HttpResponseBadRequest()

        multistatus = DAVMultistatus(self._dav_request)
        multistatus.process()

        response = HttpResponse(
            ElementTree.tostring(multistatus.xml_element), content_type="text/xml", status=207
        )
        return response

    def report(self, request, *args, **kwargs):
        return self.propfind(request, *args, **kwargs)


class DAVSingleResourceView(DAVResourceView):
    """View for single CalDAV resources."""

    def propfind(self, request, id, *args, **kwargs):  # noqa: A002
        resource = self.get_object()
        try:
            objects = resource.get_objects(request).get(pk=id)
        except resource.DoesNotExist as exc:
            raise Http404 from exc

        self._dav_request = DAVRequest(request, resource, objects)

        try:
            self._dav_request.parse()
        except SAXParseException:
            return HttpResponseBadRequest()

        multistatus = DAVMultistatus(self._dav_request)
        multistatus.process()

        response = HttpResponse(
            ElementTree.tostring(multistatus.xml_element), content_type="text/xml", status=207
        )
        return response

    def get(self, request, name, id, *args, **kwargs):  # noqa: A002
        resource: DAVResource = self.get_object()
        try:
            objs = resource.get_objects(request)
            if isinstance(objs, set):
                obj = next(filter(lambda o: o.pk == id, objs))
            else:
                obj = objs.get(pk=id)
        except resource.DoesNotExist as exc:
            raise Http404 from exc

        response = HttpResponse(content_type=resource.get_dav_content_type())
        response.write(resource.get_dav_file_content(request, objects=[obj]))
        return response


class CustomLogoutView(LogoutView):
    """Custom logout view that allows GET requests."""

    http_method_names = ["post", "get", "options"]

    def get(self, *args, **kwargs):
        return self.post(*args, **kwargs)


@method_decorator(csrf_exempt, name="dispatch")
class DAVWellKnownView(View):
    http_method_names = View.http_method_names + ["propfind", "report"]

    def get(self, request, *args, **kwargs):
        return redirect("dav_registry")

    def propfind(self, request, *args, **kwargs):
        return redirect("dav_registry")

    def post(self, request, *args, **kwargs):
        return redirect("dav_registry")
